/* 
* -cr_beg-
* ********************************************************************
* IBM Confidential
* OCO Source Materials
* 5724-U38, 5724-U39, 5724-U48, 5655-V17, 5655-V19, 5655-V35, 5655-V37
* (C) Copyright IBM Corp. 
* The source code for this program is not published or otherwise
* divested of its trade secrets, irrespective of what has
* been deposited with the U.S. Copyright Office.
* ********************************************************************
* -cr_end-
*/ 

/*
 * File: AdedStep.cpp
 *
 * Function: Main dialog class for Show Steps from extract process
 *           Display Logic and Messages derived from OS/2 DfInt.c (5/30/96)
 *           Output formatted for display in RtRichEdit Control.
 *----------------------------------------------------------------------------
 *
 * Modified:
 *
 *    $Author:   ClarkW  $
 *    $Date:   16 Apr 2004 14:12:34  $
 *    $Revision:   1.25  $
 *    $Log:   Y:/RTWIN/BSF/ARCHIVES/adedstep.cpv  $
 * 
 *    Rev 1.25   16 Apr 2004 14:12:34   ClarkW
 * UNICODE: Converted to use Generic Mapping Macros
 * 
 *    Rev 1.24   Jul 02 2001 15:06:08   KEVINM
 * Cast CString and derived class objects used in printf-like functions to
 *   const char * for the Sun platform.
 * 
 *    Rev 1.23   Jun 19 2001 13:32:32   KEVINM
 * Cast CStrings used in printf-like functions to const char * for the Sun.
 * 
 *    Rev 1.22   Jun 14 2001 09:48:12   Judyf
 * 
 * Check for GROUP in use when formatting Step1
 *
 *    Rev 1.21   Apr 17 2000 09:25:54   Judyf
 * Formatting String must have exact number of plugs.
 *
 *    Rev 1.20   Mar 07 2000 08:46:54   DAVEHEND
 *
 * Correct Version 6 warning/errors.
 *
 *    Rev 1.19   Mar 01 2000 07:14:56   Judyf
 *
 * Fix vc6.0 compiler warnings
 *
 *    Rev 1.18   Oct 14 1999 06:23:22   Judyf
 *
 * Add "Save As" to file menu
 *
 *    Rev 1.17   Sep 23 1999 09:28:02   Judyf
 * Add code to prevent indentation problem in the ShowSteps report
 *
 *    Rev 1.16   Sep 22 1999 07:07:32   Judyf
 *
 * Add spaces after periods in Show Steps messages
 *
 *    Rev 1.15   Dec 21 1998 06:35:06   JudyF
 * Fix indentation problem when formatting report
 *
 *    Rev 1.14   Oct 08 1998 15:52:38   JUDYF
 *
 * Use child table instead of parent table to determine selection criteria
 *
 *    Rev 1.13   Jul 31 1998 17:43:12   JUDYF
 *
 * Fix tabbing problem in display
 *
 *    Rev 1.12   Jul 10 1998 14:34:44   JUDYF
 *
 * Write out three-part names instead of CreatorId.Tablename
 *
 *    Rev 1.11   Jun 24 1998 13:58:58   JUDYF
 *
 * Check for untitled definition condition when formatting report
 *
 *    Rev 1.10   May 19 1998 16:32:34   JUDYF
 * Fix: DJH - 00019 (Show Steps not aligned
 *
 *    Rev 1.9   Apr 16 1998 14:11:54   SCOTR
 * Added RichEdit print capability.
 *
 *    Rev 1.8   Apr 13 1998 09:14:02   SCOTR
 * Added code to OnStepPrint() method.
 *
 *    Rev 1.7   Dec 18 1997 09:49:24   JUDYF
 *
 * Supply parent handle on constructor
 *
 *    Rev 1.6   Oct 14 1997 11:33:54   WALTERB
 *
 * Removed Pane 2 status update
 * improved speed of stuffing RichText control by pre-selecting Wait Cursor.
 *
 *    Rev 1.5   Oct 10 1997 15:22:56   WALTERB
 * Fixed problems in Cycle Step Display
 * Added Progress counter to Pane 2
 *
 *    Rev 1.4   Oct 07 1997 15:12:34   WALTERB
 * Added RichText Display to ShowSteps option
 *
 *    Rev 1.3   Sep 17 1997 08:56:12   WALTERB
 *
 * Removed unreferenced variables
 *
 *    Rev 1.2   Sep 17 1997 08:39:34   WALTERB
 * Added extra padding to display buffer for extra new lines white spaces
 *
 *    Rev 1.1   Sep 05 1997 09:55:26   WALTERB
 *
 * Lined up display of RefTbl Names, and Untraversed RelTbl Names
 *
 *    Rev 1.0   Sep 04 1997 09:43:32   WALTERB
 * Initial revision.
 *    $Author:   ClarkW  $
 *
*/
#include "RtStdAfx.h"    // Required for preprocess header

#define ADEDSTEP_CPP     // Tell include ok to export
#include "RtCommon.h"    // Common includes
#include "PswrDefs.rh"   // PswrDefs resource ids
#include "RtRdbMem.h"
#include "RtAD.h"        // Our include
#include "DXXc.h"        // AD header
#include "RtEx.h"        // For Extract data
#include "AdedDlg.h"     // Our header
#include "AdedStep.h"    // Our header
#include "RtRich.h"      // For RichText display box
#include "RtPrint.h"     // For print support
/*----------------------------------------------------------------------------
 * Copyright and fix list
 *---------------------------------------------------------------------------*/
RtCopyright();

/*----------------------------------------------------------------------------
 * AdedStepDialog message map
 *---------------------------------------------------------------------------*/
BEGIN_MESSAGE_MAP(AdedStepDialog, RtDialog)
   //{{AFX_MSG_MAP(AdedStepDialog)
        ON_BN_CLICKED(ADED_STEP_DISPLAY, OnDisplaySteps)
        ON_BN_CLICKED(ADED_STEP_PRINT, OnStepPrint)
        ON_COMMAND(ADED_MENU_SAVEAS, OnSaveAs)
        //}}AFX_MSG_MAP
END_MESSAGE_MAP()

/*----------------------------------------------------------------------------
 * Default toolbar options
 *---------------------------------------------------------------------------*/
static int BASED_CODE gToolButs[] = { -1 };

/*----------------------------------------------------------------------------
 * AdedStepDialog::AdedStepDialog
 *    Standard dialog constructor
 *---------------------------------------------------------------------------*/
AdedStepDialog::
      AdedStepDialog(RtAD *pAD, AdedDialog *Owner)
     :RtDialog(ADED_STEP_DIALOG,            // Resource id
               RTCOMP_ADED,                 // RtCompId (see RtImport.h for list)
               Owner->m_hWnd,               // Parent HWND
               gToolButs,                   // Toolbar default setup
               DefaultOption | NoModalShow)      // Dialog Options
{
   RtTraceEntry(AdedStepDialog::AdedStepDialog);
   mpAD = pAD;                         // hold onto AD object pointer
   ShowToolbar(FALSE);                 // Don't want the toolbar
   ShowErrorbar(FALSE);                // Don't want the Errbar
   ShowStatusbar(TRUE);                // Do    want the Statbar
}

/*----------------------------------------------------------------------------
 * AdedStepDialog::~AdedStepDialog
 *    Destructor
 *---------------------------------------------------------------------------*/
AdedStepDialog::
      ~AdedStepDialog()
{
   RtTraceEntry(AdedStepDialog::~AdedStepDialog);
}

/*----------------------------------------------------------------------------
 * AdedStepDialog::OnInitDialog
 *    Standard init dialog routine
 *---------------------------------------------------------------------------*/
BOOL AdedStepDialog::
      OnInitDialog()
{
   RtTraceEntry(AdedStepDialog::OnInitDialog);

   // Let RtDialog do it's work
   RtDialog::OnInitDialog();

   // Set initial caption
   CString adname = mpAD->GetADID() + _T(".") + mpAD->GetADName();
   if (adname != _T("."))
      SetCaption(adname);
   else
      SetCaption(RtString(RTRES_UNTITLED));
   mRTE.Create(ADED_STEP_DISPLAYBOX, this);

   // Set up CEdit Control to display extract results string
   Flex(FlexSize, ADED_STEP_DISPLAYBOX);

   // Send WM_COMMAND to start up logic to display output
   ShowWindow(SW_SHOW);
   PostMessage(WM_COMMAND,ADED_STEP_DISPLAY);
   return(TRUE);
}

/*----------------------------------------------------------------------------
 * AdedStepDialog::DoDataExchange
 *    Exchange variables with dialogs: display CEdit Text
 *---------------------------------------------------------------------------*/
void AdedStepDialog::
      DoDataExchange(CDataExchange* pDX)
{
   // Don't trace high-velocity methods

   RtDialog::DoDataExchange(pDX);

   //{{AFX_DATA_MAP(AdedStepDialog)
        //}}AFX_DATA_MAP
}

/*----------------------------------------------------------------------------
 * AdedStepDialog::CanClose (overriden from RtDialog)
 *    See if we can close this window just now
 *---------------------------------------------------------------------------*/
BOOL AdedStepDialog::
      CanClose(CloseReasons Reason)
{
   RtTraceEntry(AdedStepDialog::CanClose);
   if (!RtDialog::CanClose(Reason))
      return(FALSE);
   return(TRUE);
}

/*----------------------------------------------------------------------------
 * AdedStepDialog::OnSaveAs()
 *    Save Show Steps report to a file
 *--------------------------------------------------------------------------*/
void AdedStepDialog::OnSaveAs()
{
   RtTraceEntry(CreaReviewDialog::OnMenuSaveAs);

   RtWaitCursor wait(this);

   //  Go save the file as a text (.TXT) file
   mRTE.SaveAs(NULL,
               RT_OUT_DEFAULT_EXT,
               RT_OUT_DEFAULT_FILTER);
}

/*----------------------------------------------------------------------------
 * AdedStepDialog::OnStepPrint()
 *    Core of Show Steps Processing
 *---------------------------------------------------------------------------*/
void AdedStepDialog::OnStepPrint()
{
   RtPrint printer;

   RtTraceEntry(AdedStepDialog::OnStepPrint);

   if (printer.DoPrintRichEdit (&mRTE, this) == FALSE) {
       RtPopup (printer.mError, RTPOPUP_BUTTON_KEY_OKDEFAULT, RTPOPUP_ERROR);
   }
}

/*----------------------------------------------------------------------------
 * AdedStepDialog::OnDisplaySteps()
 *    Core of Show Steps Processing
 *---------------------------------------------------------------------------*/
void AdedStepDialog::OnDisplaySteps()
{
   RtTraceEntry(AdedStepDialog::OnDisplaySteps);

   MEM_HEAP_HANDLE mh = NULL;
   UINT    ErrorResid;
   CString PortPlugs;
   CString ErrorDisplay;

   DXPL_MAIN * pDXPLMain;
   BOOL bOK;
   CString csLabel;

   csLabel = mpAD->GetADID();
   csLabel += _T(".");
   csLabel += mpAD->GetADName();

   //------------------------------------------------Initialize display
   ErrorDisplay.Empty();  // Initialize Strings

   //---------------------------------------Get Show Steps data from NEC
// SetCaption(csLabel);
   bOK = mpAD->GetShowStepsData( pDXPLMain, &mh, ErrorResid, PortPlugs);

   // Call routine to format it for CEdit display, or show error
   if (bOK)
      bOK = FormatStepsData((void *)pDXPLMain, csLabel);
   else
      ErrorDisplay.Format(_T("%s\r\n%s\r\n"),
                          (const TCHAR *) RtString(ADED_STEP_ERROR),
                          (const TCHAR *) RtMergeMsg(ErrorResid,PortPlugs));

   if (!bOK)   // if error, display, and close out
   {
      RtPopup(ErrorDisplay, RTPOPUP_BUTTON_KEY_OKDEFAULT, RTPOPUP_ERROR);
      if (mh)
         Mem_HeapClose(mh);
      PostMessage(WM_COMMAND, IDCANCEL);  // close out
      return;
   }

   if (mh)
      Mem_HeapClose(mh);
}
/*----------------------------------------------------------------------------
 * AdedStepDialog::FormatStepsData()
 *    Convert Show Steps Data to Display data for Edit Window
 *    Returns TRUE if successful
 *---------------------------------------------------------------------------*/
#define DX_USED _T('U')  // used to mark steps traversed
BOOL AdedStepDialog::FormatStepsData(void *pDXPL, CString ADLabel)
{
   short  StepIndex, i;
   short  FirstLine, FirstStep, FirstInstance, FirstPath;
   short  NumEntryPath, RefFlag;
   float Indent0 = 0.0;  // indent position, in inches
   float Indent1 = (float).4;
// float Indent2 = 1.0;

   CString csPctDone;
   CString TableName;
   CString CreatorID;
   CString DBAlias;
   CString FullTblName;
   CString CompleteName;

   _TUCHAR *LastComma;
   _TUCHAR TempString1[80], TempString2[80];
   DXPL_TABLE *       pTable;
   DXPL_MAIN *        pMain;
   DXPL_STEP *        pStep;
   DXPL_TABLE *       pTbl;
   DX_INPUT_TABLE *   pTestTbl;
   DXPL_TRAV_PATH *   pEPath;
   DXPL_TRAV_PATH *   pPath;
   DXPL_CYCLE_PATH *  pCPath;
   DXPL_UNTRAV_REL *  pUTRel;
   BOOL bSelCrit;

   RtTraceEntry(AdedStepDialog::FormatStepsData);

   // Initialize
   pMain         = (DXPL_MAIN *) pDXPL;
   StepIndex     = 0;
   FirstStep     = TRUE;
   FirstInstance = TRUE;
   RefFlag       = FALSE;

   /*----------------------------------------------------------------------*/
   /* PREPARE THE RICH TEXT DISPLAY BUFFER:                                */
   /*   Set up formatting, list out title                                  */
   /*----------------------------------------------------------------------*/

// RtLockWindow* pLock = new RtLockWindow(mRTE); // prevent flashing of display
// pLock->Lock();
   RtLockWindow lock(mRTE,TRUE);
// mRTE.Clear(TRUE);
   mRTE.SetWordwrap(RtRichEdit::WrapFixed, 20);

   // Set up title
   mRTE.SetColor(RtRichEdit::Black);
   mRTE.SetBold();
   mRTE.AlignPara(RtRichEdit::ParaCenter);
   mRTE.SetFont(RtRichEdit::MSSansSerif, 12);
   if (ADLabel == _T("."))
      mRTE.AddText( RtString(ADED_STEP_TITLE_UNTITLED) );
   else
      mRTE.AddText( RtFormatMsg(ADED_STEP_TITLE, _T("%s"), (const TCHAR *) ADLabel) );
   mRTE.AddText( _T("\n")); // must guarantee endline to end 1st 'paragraph'

   // Set up normal print line
   mRTE.SetFont(RtRichEdit::MSSansSerif, 10);
   mRTE.SetBold(FALSE);
   mRTE.AlignPara(RtRichEdit::ParaLeft);
   mRTE.AddText(_T("\n"));


   /*----------------------------------------------------------------------*/
   /* BEGIN TO POPULATE THE DISPLAY BUFFER:                                */
   /*   Search the whole STEP list                                         */
   /*----------------------------------------------------------------------*/

// BeginWaitCursor(); // display the hourglass cursor
   RtWaitCursor wait(this);
   SetStatus(ADED_STEP_FORMATTING);

   for (pStep = pMain->p1stStep ; //=============================STEP LOOP
        pStep != NULL ; pStep = pStep->pNext)
   {
      StepIndex++;  // Increase the Step Counter

      FirstLine = TRUE;
      mRTE.LeftIndentPara(5.0);
      mRTE.AddText(_T(" \n"));
      mRTE.LeftIndentPara(Indent0);
      mRTE.SetBold();
      mRTE.AddText(RtFormatMsg(ADED_STEP_STEPNUM, _T("%2d"), StepIndex) ); //"\nStep ^1: "
      mRTE.AddText(_T("\n"));
      mRTE.SetBold(FALSE);
      mRTE.LeftIndentPara(Indent1);
      /*--------------------------------------------------------------------*/
      /* Note: The following newline was added to prevent an RichEdit       */
      /*         bug from occuring where the Step was not always unindented.*/
      /*         Apparently, the newline at the end of a paragraph          */
      /*         determines the formatting for the entire paragraph.        */
      /*         CRichEdit has problems when the text of the paragraph is   */
      /*         long.     (Clark has researched this issue)                */
      /*--------------------------------------------------------------------*/
      mRTE.AddText(_T("\n"));

      /*-------------------------------------------------------------------*/
      /* Process Each Step: either as cycle, or non-cycle step             */
      /*-------------------------------------------------------------------*/
//    mRTE.LeftIndentPara(Indent1); // indent steps
     if (pStep->CycleCell != TRUE) //----------------- NON-cycle  Step
      {
         if (FirstStep) //--------------If it's first step
            {
            // Get the full name from ADTbl
            mpAD->GetCompleteStrt(&DBAlias, &CreatorID, &TableName);
            mpAD->GetCompleteStrt(CompleteName);

            // Search the whole table list, to find Start Table's
            // pointer. In Dummy situation, p1stTbl->pADTbl is NULL
            // We can't get SelCrit's info from this pointer.
            for (pTable = pStep->p1stTbl ; pTable != NULL ;
                 pTable = pTable->pNext)
            {
               pTestTbl = pTable->pInputTbl; // get next link

               if (DBAlias == (const TCHAR *)pTable->pInputTbl->pDBAliasName &&
                   CreatorID == (const TCHAR *)pTable->pInputTbl->pCreatorID &&
                   TableName == (const TCHAR *)pTable->pInputTbl->pCreatorID)
                   break;
            }
            // find if any columns in this table has selection criteria
            FullTblName = CompleteName;
            bSelCrit = TblHasSelCrit(CompleteName);

            // "Extract Rows from Start Table ^1. "
            mRTE.AddText(RtFormatMsg(ADED_STEP_STARTTBL, _T("%s"), (const TCHAR *) CompleteName) );

            // There is a Point-and-Shoot file
            if (mpAD->GetPnSState() != DsxAD::None)
            {
               // "Row List is used and Determines the Rows Selected."
            mRTE.AddText( RtString(ADED_STEP_START_PNS) );
            }
            else if ( bSelCrit || pTestTbl->EveryNth != 0  ||
                      pTestTbl->RowLimit != 0 || mpAD->GetNofMKeyColumn().GetLength() > 0)
            {
               // "Selection Criteria and/or Statistical Controls are used,"
               // "these Determine the Rows Selected."
               mRTE.AddText( RtString(ADED_STEP_START_SEL) );
            }

            else
            {
               //"No Row List, Selection Criteria or Statistical Controls used, therefore "
               //"Start Table does not need to be Revisited, even if part of a Cycle."
               mRTE.AddText( RtString(ADED_STEP_NOSEL1));
               mRTE.AddText( RtString(ADED_STEP_NOSEL2));
            }

            FirstStep = FALSE;  // Finished the first step

            // Set a blank line for separate two parts
            mRTE.AddText(_T("\n"));

            continue;  // Continue the 1st FOR loop
         } //------------ End of First Step if

         // Initial the first instance
         FirstInstance = TRUE;

         // After finishing the first step, we search the whole p1stEntryPath
         for (pEPath = pStep->p1stEntryPath ; //--------------Go down Entrypath
               pEPath != NULL ; pEPath = pEPath->pNext)
         {
            // If this Entry Path is used
           if (pEPath->PathType == DX_USED)
               continue;

            // "in Step ^1"
            _tcscpy((TCHAR *) TempString1, RtFormatMsg(ADED_STEP_INSTEP, _T("%2d"), pEPath->StepNum + 1));

            // Search the rest of EntryPath List, to find whether or not
            //   there are some other Paths with same Relationship
            FirstPath = TRUE;
            for (pPath = pEPath->pNext ; pPath != NULL ;
                                         pPath = pPath->pNext)
            {
               // There is a EntryPath related to same REL
               if (pPath->pInputRel == pEPath->pInputRel) {
                  _stprintf((TCHAR *)TempString2, _T(", %d"), pPath->StepNum + 1);

                  // This Path has been used
                  pPath->PathType = DX_USED;

                  if (FirstPath)
                  {
                     _stprintf((TCHAR *)TempString1, _T("in Steps %d"), pEPath->StepNum + 1);
                     FirstPath = FALSE;
                  }

                  _tcscat((TCHAR *)TempString1, (TCHAR *)TempString2);
               }
            }

            // Change "in Steps X, Y, Z" to "in Steps X, Y and Z"
            if ((LastComma = (_TUCHAR *)_tcsrchr((TCHAR *)TempString1, _T(','))) != NULL)
            {
               _tcscpy((TCHAR *)TempString2, (TCHAR *)LastComma + 1);
               _tcscpy((TCHAR *)LastComma, RtString(ADED_STEP_AND));
               _tcscat((TCHAR *)TempString1, (TCHAR *)TempString2);
            }

            if (pEPath->PathType == DX_CHILD_PATH) //-------process child path
            {
               // find if any columns in this table has selection criteria
               CString name = pEPath->pInputRel->pChlDBAliasName;
               name += _T(".");
               name += pEPath->pInputRel->pChlCreatorID;
               name += _T(".");
               name += pEPath->pInputRel->pChlTableName;
               bSelCrit = TblHasSelCrit(name);

                   //  "Extract Rows from ^1.^2.^3 which are Children of Rows "
               mRTE.AddText(RtFormatMsg(ADED_STEP_CHROW_PREV1, _T("%s%s%s"),
                       pEPath->pInputRel->pChlDBAliasName,
                       pEPath->pInputRel->pChlCreatorID,
                       pEPath->pInputRel->pChlTableName) );
                   //  "Previously Extracted from ^1.^2.^3 ^4 using Relationship ^5.",
               mRTE.AddText(RtFormatMsg(ADED_STEP_CHROW_PREV2, _T("%s%s%s%s%s"),
                     pEPath->pInputRel->pParDBAliasName,
                     pEPath->pInputRel->pParCreatorID,
                     pEPath->pInputRel->pParTableName,
                     (const TCHAR *) TempString1,
                     pEPath->pInputRel->pRelName) );

               if (bSelCrit)   // There are Selection Criteria
                  // "Limited by Selection Criteria."
                  mRTE.AddText( RtString(ADED_STEP_CHROW_PREVLIM) );
               else
               {
                  mRTE.AddText( _T(". "));
               }
            }     // ---------------------------- End of DX_CHILD_PATH
            else  // ---------------------------- Process Parent path DX_PAR_PATH
            {
               // "Extract Rows from ^1.^2.^3 which are Parents of Rows "
               mRTE.AddText( RtFormatMsg(ADED_STEP_PAROW_PREV1, _T("%s%s%s"),
                         pEPath->pInputRel->pParDBAliasName,
                         pEPath->pInputRel->pParCreatorID,
                         pEPath->pInputRel->pParTableName) );
               // "Previously Extracted from ^1.^2.^3 ^4 to satisfy an RI rule using Relationship ^5.",
               mRTE.AddText( RtFormatMsg(ADED_STEP_PAROW_PREV2,  _T("%s%s%s%s%s"),
                        pEPath->pInputRel->pChlDBAliasName,
                        pEPath->pInputRel->pChlCreatorID,
                        pEPath->pInputRel->pChlTableName,
                        (const TCHAR *) TempString1,
                        pEPath->pInputRel->pRelName ) );
            }  //----------------------------- End of process Parent

            if (!FirstInstance)
            {
               mRTE.AddText(_T("\n\t-- AND --\n"));
            }
         }   // ---------------End of the p1stEntryPath FOR loop

         // Set a blank line for separate two parts
         mRTE.AddText(_T("\n"));
      }      //------------------------- End of the Non-Cycle Step if block

      /*-------------------------------------------------------------------*/
      /*   Process a Cycle                                                 */
      /*-------------------------------------------------------------------*/
      else  // -------------------- This Step is a Cycle
      {
         // Build the first part of the sentence:  "Table(s)"

         mRTE.LeftIndentPara(Indent1);  // INDENT THESE line(s):
         if (pStep->NumTbls > 1)
            mRTE.AddText( RtString(ADED_STEP_TABLES) );
         else
            mRTE.AddText( RtString(ADED_STEP_TABLE) );

         // Search the whole p1stTbl list, build the TABLE NAME's part
         for (pTbl = pStep->p1stTbl ; pTbl != NULL ; pTbl = pTbl->pNext)
         {
            pTestTbl = pTbl->pInputTbl;
            if(pTestTbl->pDBAliasName[0] != _T('\0'))
               mRTE.AddText((const TCHAR*) pTestTbl->pDBAliasName);
            else
               mRTE.AddText((const TCHAR*)mpAD->GetDefQualifierDBA());
            mRTE.AddText(_T("."));
            if(pTestTbl->pCreatorID[0] != _T('\0'))
               mRTE.AddText((const TCHAR*) pTestTbl->pCreatorID);
            else
               mRTE.AddText((const TCHAR*)mpAD->GetDefQualifierCID());
            mRTE.AddText(_T("."));
            mRTE.AddText((TCHAR *)pTestTbl->pTableName);
            mRTE.AddText( _T(", "));
         }

         if (pStep->NumCyclePaths > 1)
            mRTE.AddText( RtString(ADED_STEP_USERELS)); // "using Relationships ");
         else
            mRTE.AddText( RtString(ADED_STEP_USEREL)); // "using Relationship ");

         // Search the whole p1stCyclePath, build the REL NAME's part
         for (pCPath = pStep->p1stCyclePath ; pCPath != NULL ;
                                              pCPath = pCPath->pNext)
         {
            mRTE.AddText((TCHAR *)pCPath->pInputRel->pRelName);
            mRTE.AddText( _T(", "));
         }

         // Dummy situation, the Start Table is in the Cycle, so
         // there is no Entry Path to get in this Cycle. We build
         // other texts.
         if (pStep->p1stEntryPath->pInputRel == NULL)
         {
            if (pStep->NumTbls > 1)
               mRTE.AddText( RtString(ADED_STEP_FORM) ); //"form"
            else
               mRTE.AddText( RtString(ADED_STEP_FORMS) ); //"forms"

            //" a Cycle.  This Cycle includes the Start Table ^1.^2. "
            mRTE.AddText( RtFormatMsg(ADED_STEP_CYCLEINC,  _T("%s"), (const TCHAR *) CompleteName));
            goto  FetchText;
         }

         NumEntryPath = 0;

         // Do prescan, to find the real num. of the Entry Paths
         for (pEPath = pStep->p1stEntryPath ; pEPath != NULL ;
                                              pEPath = pEPath->pNext)
         {
            if (pEPath->PathType == DX_USED)
               continue;
            NumEntryPath++;
            for (pPath = pEPath->pNext ; pPath != NULL ; pPath = pPath->pNext)
               if (pPath->pInputRel->pRelName == pEPath->pInputRel->pRelName)
                  pPath->PathType = DX_USED;
         }

         // Build the last part of the sentence
         if (pStep->NumTbls > 1)
            mRTE.AddText( RtString(ADED_STEP_FORM) ); //"form"
         else
            mRTE.AddText( RtString(ADED_STEP_FORMS) ); //"forms"

          //  " a Cycle.  This Cycle has"
         mRTE.AddText( RtString(ADED_STEP_CYCLEHAS) );

         _stprintf((TCHAR *) TempString2, _T(" %d "), NumEntryPath);
         mRTE.AddText((const TCHAR*)TempString2);

         if (NumEntryPath == 1) // "Input Path which Processes as Follows:"
            mRTE.AddText( RtString(ADED_STEP_INPPATH1));
         else  //"Input Paths which Process as Follows:"
            mRTE.AddText( RtString(ADED_STEP_INPPATH2) );

FetchText:
         // Set a blank line for separate two parts
         mRTE.AddText(_T("\n"));

         // Search the whole p1stEntryPath
         for (pEPath = pStep->p1stEntryPath ; pEPath != NULL ;
                                              pEPath = pEPath->pNext)
         {
            // Dummy situation, we don't need this part of text
            if (pStep->p1stEntryPath->pInputRel == NULL)
               break;

            // If this Entry Path is used
            if (pEPath->PathType == DX_USED)
               continue;

            // "in Step ^1"
            _tcscpy((TCHAR *) TempString1,RtFormatMsg(ADED_STEP_INSTEP, _T("%2d"),
                                                   pEPath->StepNum + 1) );

            FirstPath = TRUE;

            // Search the rest of EntryPath List, to find whether or not
            //  there are some other Paths with same Relationship
            for (pPath = pEPath->pNext ; pPath != NULL ;
                                         pPath = pPath->pNext)
            {
               // There is a EntryPath related to same REL
               if (pPath->pInputRel->pRelName == pEPath->pInputRel->pRelName)
               {
                  _stprintf((TCHAR *) TempString2, _T(", %d"), pPath->StepNum + 1);

                  // This Path has been used
                  pPath->PathType = DX_USED;

                  if (FirstPath)
                  {
                     _tcscpy((TCHAR *) TempString1, RtFormatMsg(ADED_STEP_INSTEPS,  _T("%d"), pEPath->StepNum + 1) );
                     FirstPath = FALSE;
                  }

                  _tcscat((TCHAR *) TempString1, (TCHAR *)TempString2);
               }
            } //--------------End of for paths loop

            // Change "in Steps X, Y, Z" to "in Steps X, Y and Z"
            if ((LastComma = (_TUCHAR *)_tcsrchr((TCHAR *)TempString1, _T(','))) != NULL)
            {
               _tcscpy((TCHAR *) TempString2, (TCHAR *)LastComma + 1);
               _tcscpy((TCHAR *) LastComma, RtString(ADED_STEP_AND));
               _tcscat((TCHAR *) TempString1, (TCHAR *)TempString2);
            }


            if (pEPath->PathType == DX_CHILD_PATH)
            {
               // find if any columns in this table has selection criteria
               FullTblName =
                  pEPath->pInputRel->pChlDBAliasName;
               if (!FullTblName.IsEmpty())
                  FullTblName += _T(".");
               FullTblName += pEPath->pInputRel->pChlCreatorID;
               FullTblName += _T(".");
               FullTblName += pEPath->pInputRel->pChlTableName;
               bSelCrit = TblHasSelCrit(FullTblName);

               //"Extract Rows from %s.%s which are Children of Rows "
               mRTE.AddText( RtFormatMsg(ADED_STEP_CHROW_PREV1, _T("%s%s%s"),
                             pEPath->pInputRel->pChlDBAliasName,
                             pEPath->pInputRel->pChlCreatorID,
                             pEPath->pInputRel->pChlTableName) );

               // "Previously Extracted from %s.%s %s using Relationship %s ",
               mRTE.AddText( RtFormatMsg(ADED_STEP_CHROW_PREV2, _T("%s%s%s%s%s"),
                             pEPath->pInputRel->pParDBAliasName,
                             pEPath->pInputRel->pParCreatorID,
                             pEPath->pInputRel->pParTableName,
                             (const TCHAR *) TempString1,
                             pEPath->pInputRel->pRelName) );

               if (!bSelCrit)  // There are not any Selection Criteria
               {
                  mRTE.AddText(_T(". "));
               }
               else // " Limited by Selection Criteria."
               {
                  mRTE.AddText( RtString(ADED_STEP_CHROW_PREVLIM) );
               }

            } // End of DX_CHILD_PATH

            else   // DX_PAR_PATH
            {
               // "Extract Rows from ^1.^2 which are Parents of Rows "
               mRTE.AddText( RtFormatMsg(ADED_STEP_PAROW_PREV1,  _T("%s%s%s"),
                             pEPath->pInputRel->pParDBAliasName,
                             pEPath->pInputRel->pParCreatorID,
                             pEPath->pInputRel->pParTableName) );

               // "Previously Extracted from ^1.^2 ^3 to satisfy an RI "
               // "rule using Relationship ^4.",
               mRTE.AddText(RtFormatMsg(ADED_STEP_PAROW_PREV2, _T("%s%s%s%s%s"),
                            pEPath->pInputRel->pChlDBAliasName,
                            pEPath->pInputRel->pChlCreatorID,
                            pEPath->pInputRel->pChlTableName,
                            (const TCHAR *) TempString1,
                            pEPath->pInputRel->pRelName) );
            }  // End of DX_PAR_PATH

            // Set a blank line for separate two parts
            mRTE.AddText(_T("\n")); //
         } // End of the p1stEntryPath FOR loop

         mRTE.LeftIndentPara(Indent1); // back to previous indentation
      } //------------------- End of the Cycle Step Operation

   } // ==================================End of the Step List FOR loop

   /*----------------------------------------------------------------------*/
   /* If there are some Untraversed Tables in this Extract                 */
   /*----------------------------------------------------------------------*/
   mRTE.AlignPara(RtRichEdit::ParaLeft);         // Back to normal indentation

   if ((pTbl = pMain->p1stUnTravTbl) != NULL)
      {
      FirstLine = TRUE;

      // Search the whole UnTravTbls list
      for ( ; pTbl != NULL ; pTbl = pTbl->pNext)
      {
         // Build margin
         if (FirstLine)
         {              //  "Untraversed Table(s):       "
            mRTE.LeftIndentPara(Indent0);
            mRTE.SetBold();
            mRTE.AddText( _T("\n"));
            mRTE.AddText( RtString(ADED_STEP_UNTRAVTBL));
            mRTE.AddText( _T("\n"));
            mRTE.SetBold(FALSE);
            mRTE.LeftIndentPara(Indent1); // indent steps
//          mRTE.SetParaTabs(Indent1);
            FirstLine = FALSE;
         }
         // Build Untraversed Table Name
         CString tmpName;
         if (pTbl->pInputTbl->pDBAliasName[0] == _T('\0'))
            tmpName = mpAD->GetDefQualifierDBA();
         else
            tmpName =  (TCHAR *)pTbl->pInputTbl->pDBAliasName;
         tmpName += _T(".");
         if (pTbl->pInputTbl->pCreatorID[0] == _T('\0'))
            tmpName += mpAD->GetDefQualifierCID();
         else
            tmpName +=  (TCHAR *)pTbl->pInputTbl->pCreatorID;
         tmpName += _T(".");
         tmpName +=  (TCHAR *) pTbl->pInputTbl->pTableName;
         mRTE.AddText( tmpName);
         mRTE.AddText( _T("\n"));;
      }
      mRTE.AddText( _T("\n"));
      mRTE.LeftIndentPara(Indent0); // indent steps
   }

   /*----------------------------------------------------------------------*/
   /* If there are some Untraversed Relationships in this Extract */
   /*----------------------------------------------------------------------*/
   if ((pUTRel = pMain->p1stUnTravRel) != NULL)
   {
      FirstLine = TRUE;

      // Search the whole UnTravRels list
      for ( ; pUTRel != NULL ; pUTRel = pUTRel->pNext)
      {
         // Build margin
         if (FirstLine)
         {  //"Untraversed Relationship(s):"
            mRTE.SetBold();
            mRTE.LeftIndentPara(Indent0);
            mRTE.AddText( _T("\n") );
            mRTE.AddText( RtString(ADED_STEP_UNTRAVREL) );
            mRTE.AddText( _T("\n") );
            mRTE.LeftIndentPara(Indent1); // indent steps
            mRTE.SetBold(FALSE);
            FirstLine = FALSE;
         }

         // Build Untraversed Relationship Name
         mRTE.AddText( (TCHAR *) pUTRel->pInputRel->pRelName);
         mRTE.AddText( _T("\n"));
      }

      // Set a blank line
      mRTE.AddText( _T("\n"));;
      mRTE.LeftIndentPara(Indent0); // indent steps
   }

   /*----------------------------------------------------------------------*/
   /* If there are some Reference Tables is this AD */
   /*----------------------------------------------------------------------*/

   DsxADTblCursor TblCursor(*mpAD);

   FirstLine = TRUE;
   for (i=1, TblCursor.First();    TblCursor.Valid();    i++, TblCursor.Next())
   {
     // Refresh Table data for both source and target table
      mpAD->GetTblsCompleteSrc(TblCursor, FullTblName);

      RefFlag = (mpAD->GetTblsRefTable(TblCursor)) ? TRUE : FALSE;

      if (!RefFlag)
         continue;   // This table is not, check next

         // Build margin indent
      if (FirstLine)
      {
         // "Reference Table(s):       "
        mRTE.SetBold();
        mRTE.LeftIndentPara(Indent0);
        mRTE.AddText( _T("\n"));
        mRTE.AddText( RtString(ADED_STEP_REVTAB) );
        mRTE.AddText( _T("\n"));;
        mRTE.SetBold(FALSE);
        FirstLine = FALSE;
      }
      mRTE.LeftIndentPara(Indent1);
      mRTE.AddText(FullTblName);
      mRTE.AddText( _T("\n"));;

      // If this reference table has selection criteria
      bSelCrit = TblHasSelCrit(FullTblName);
      mRTE.LeftIndentPara((float).8);
      if (bSelCrit)
          mRTE.AddText( RtString(ADED_STEP_SELCRIT) ); //"Using Selection Criteria"
      else
          mRTE.AddText( RtString(ADED_STEP_UNLIMIT)); //"All Rows                 "

      mRTE.AddText( _T("\n"));;
   }

   if (RefFlag == TRUE)
   {
      // Set a blank line
      mRTE.AddText( _T("\n"));;
   }
   /*----------------------------------------------------------------------*/
   /*----------------------------------------------------------------------*/

   mRTE.SetWordwrap(RtRichEdit::WrapWindow);  // allow dynamic display resize

   CHARRANGE sel; // reset cursor at top, with nothing selected
   sel.cpMin = sel.cpMax = 0;
   mRTE.SetSel(sel);

   // To investigate RT display contents:
   // CString outp = mRTE.GetRtfText(); FILE *fs=fopen("step.txt","w");fprintf(fs,(LPCSTR)outp);fclose(fs); //!!

// pLock->Unlock();
// delete pLock;
// EndWaitCursor();  // end display of the hourglass cursor
   SetStatus();      // clear status display

   return(TRUE);
}
/*----------------------------------------------------------------------*
 *  AdedStepDialog::TblHasSelCrit(CString FullTblName)
 *    returns true or false as to whether table has selection criteria
 *----------------------------------------------------------------------*/
BOOL AdedStepDialog::TblHasSelCrit(CString FullTblName)
{
   DsxADTbls * tbl = mpAD->FindTblsBySrcTable(FullTblName);

   if (!tbl) return(FALSE);

   // If table has an SQL clause associated with it, return TRUE.
   if (mpAD->GetTblsSQLText(tbl) != _T(""))
       return(TRUE);

   // Determine if atleast one column has selection criteria
   DsxADColCursor RowCursor(*tbl);
   BOOL bSelCrit = FALSE;
   forCursor(RowCursor)
   {
      // is there Selection Criteria for this col?
      if (!mpAD->GetColsSelCrit(RowCursor).IsEmpty())
      {
         bSelCrit = TRUE;
         break;
      }
   }
   return (bSelCrit);
}
