/*
* -cr_beg-
* ********************************************************************
* IBM Confidential
* OCO Source Materials
* 5724-U38, 5724-U39, 5724-U48, 5655-V17, 5655-V19, 5655-V35, 5655-V37
* (C) Copyright IBM Corp. 1997,2009,2010
* The source code for this program is not published or otherwise
* divested of its trade secrets, irrespective of what has
* been deposited with the U.S. Copyright Office.
* ********************************************************************
* -cr_end-
*/

#ifndef DXXCINT_H
#define DXXCINT_H
/*--------------------------------------------------------------------*
 *                                                                    *
 * HEADER FILE NAME = DXXCINT                                         *
 *                                                                    *
 * CLASSIFICATION = PRIVATE file internal to Component.               *
 *                                                                    *
 * HEAD FILE TYPE = Component Private Interface declaration           *
 *                                                                    *
 * STRUCTURE DEFINES =  none                                          *
 *                                                                    *
 * PRODUCT = Non-External Components                                  *
 *                                                                    *
 * COMPONENT = Data Extract                                           *
 *                                                                    *
 * DESCRIPTIVE NAME = Relational Extract Engine                       *
 *                                                                    *
 *                                                                    *
 * VERSION = 1.0                                                      *
 *                                                                    *
 * USAGE = This header file must be included in internal DXX          *
 *         source files.                                              *
 *                                                                    *
 *         Information in this block is private to Data Extract       *
 *         In general other components should refrain from            *
 *         referencing this header files.                             *
 *                                                                    *
 *--------------------------------------------------------------------*
 *                                                                    *
 * NOTES :                                                            *
 *    DEPENDENCIES = NEC Framework Specification                      *
 *    RESTRICTIONS = None                                             *
 *    REGISTER CONVENTIONS = Standard for the platform                *
 *    PATCH LABEL = N/A                                               *
 *    PROCESSOR = Microsoft Visual C++, IBM Visual Age C++, SAS/C     *
 *    ATTRIBUTES = Read only, header file                             *
 *                                                                    *
 *    AUTHOR = D. HOESCHELE                                           *
 *                                                                    *
 * CHANGE ACTIVITY :                                                  *
 *   Vers  ID     Date    Change Description                          *
 *    1.0  DH   02/14/97  Initial creation                            *
 *                                                                    *
 *--------------------------------------------------------------------*
 * RTC/Jazz change log
 *
 * 13 Jul 2011 osheat 8.1.0 4033 WI 28046
 * Teradata access reporting.
 *
 * 27 May 2011 ShilpaJ 8.1.0 RTC 20713.
 * Added for Teradata Support DXX  DATA EXTRACT
 *
 * 10 April 2011 SagarB 8.1.0 RTC 22144.
 * Uplift the Maximum Row limit supported in Optim
 * Phase 1 - Limited to Archive/Extract,insert,Delete,convert and Load.
 *
 * 17 Jun 2010 SagarB 7.3.0 WI #3682
 * Changes to incorporate Compression Enhancement
 *
 * 31 Aug 2009 rtandon1 7.2.0 3012
 * Changes for Long Computer Name
 *
 *
 * Modified:
 *
 *    $Author:   BOBO  $
 *    $Date:   20 Apr 2009 12:49:22  $
 *    $Revision:   1.3  $
 *    $Log:   Y:\RTWIN\711\ARCHIVES\DXXCINT.H_v  $
 *
 *    Rev 1.3   20 Apr 2009 12:49:22   BOBO
 * - Add bAsChar bit to DX_COL_ATTR_LIST
 * Opt711: 2327   IR 72718
 *
 *    Rev 1.2   22 Oct 2008 16:31:18   OPTIMBLD
 * Remove keyword flagged by Source Code Scanning Utility "S O F T E C H"
 *
 *    Rev 1.1   21 Oct 2008 09:48:00   OPTIMBLD
 * Remove keyword flagged by Source Code Scanning Utility "B A S E D  O N"
 *
 *    Rev 1.0   07 Apr 2008 08:15:54   BOBO
 * Initial revision.
 *
 *--------------------------------------------------------------------*/

#ifdef PST_CPLUS          /* Don't mangle names */
extern "C" {
#endif

#include <ENVCFATT.H>  // File attribute defs

/* Some defines */
#define PST_VOLATILE volatile



/* some constants internal to DX  */
#define DX_UNKNOWN               '\0'
#define DX_CYCLE_PATH            'C'
#define DX_NOT_CYCLE_PATH        'N'
#define DX_USER_STOP              1
#define DXX_ROUNDTRIP_ERROR       1

#define DX_ONLINE_NOTIFY_COUNT    100
#define DX_BATCH_NOTIFY_COUNT    5000

#define      MAX_IN_COLS       50
#define      MAX_AND_OR_COLS   10

/* KSB - 03/17/2011 - RTC-22144
   Uplift the Maximum Row limit supported in Optim.
   Phase 1 - Limited to Archive, Insert, Delete, Convert and Load.*/
#define DX_MAXROWS_PER_TABLE   2000000000

/* values for PKType in the DX_PK struct */
#define DX_PRIMARY        0
#define DX_EXTENDED       1  /* DX_PK is for an extended Rel or
                                a base table PK for a Joined view */

/* Values to help decide which cols to fetch and Put */
#define DX_GET_COL 1
#define DX_PUT_COL 2


/* define for how many SELECT stmts with COL IS NULL clauses.  This
   number should correspond to the number of dynamic cursors that
   the DSQL subsystem manages. */
#define  NUM_NULL_STMTS   16




/* General purpose macros to test, set, and clear bit flags */
#define DX_CLR_FLAG(fld,flg)  ((fld) &= ~(flg))
#define DX_SET_FLAG(fld,flg)  ((fld) |=  (flg))
#define DX_TST_FLAG(fld,flg)  ((fld) &   (flg))



#define  DX_CHECK_STOP(pCB)                                            \
{   if (pCB->ErrCode) {                                                \
        PST_GOTO_RETURN_ERR(pThrd, Lbl, DX_STOP);                      \
    }                                                                  \
    if (pCB->pAbortSemaphore != NULL && *pCB->pAbortSemaphore != 0) {  \
        pCB->ErrCode = DX_CANCEL;                                      \
        PST_GOTO_RETURN_ERR(pThrd, Lbl, DX_STOP);                      \
}   }


#define DXX_START_DSQL_TIMING                            \
    if (TRACE_TEST(pThrd, SL31)) {                       \
        TimingsOn = TRUE;                                \
        RDB_TraceTimings(RDB_TIMING_INIT);               \
        pRDBThrdCB->Timings.TimingsActive = TRUE;        \
        _ftime(&Time0);                                  \
    }

#define DXX_START_DSQL_TIMING_WITH_STATS                 \
    if (TRACE_TEST(pThrd, SL31) || pDXCB->BuildStats) {  \
        TimingsOn = TRUE;                                \
        RDB_TraceTimings(RDB_TIMING_INIT);               \
        pRDBThrdCB->Timings.TimingsActive = TRUE;        \
        pDXCB->SaveRowsPut = pDXCB->TotalRowsPut;        \
        _ftime(&Time0);                                  \
    }


#define DXX_START_DSQL_TIMING_WITH_STATS1(pWA)           \
    if (TRACE_TEST(pThrd, SL31) || pDXCB->BuildStats) {  \
        TimingsOn = TRUE;                                \
        RDB_TraceTimings(RDB_TIMING_INIT);               \
        pRDBThrdCB->Timings.TimingsActive = TRUE;        \
        pWA->CntRowsPut=0;                               \
        _ftime(&Time0);                                  \
    }



#define DXX_END_DSQL_TIMING(str)                                   \
    if (TimingsOn) {                                               \
        _ftime(&Time1);                                            \
        TimeDif =  (Time1.time - Time0.time) * 1000                \
             + (Time1.millitm - Time0.millitm);                    \
        ETRACE2(pThrd, "%.1900s; TotTime=%ld MilliSec", str, TimeDif);  \
        RDB_TraceTimings(RDB_TIMING_TRACE);                        \
        memset(&(pRDBThrdCB->Timings), '\0',                       \
               sizeof(pRDBThrdCB->Timings));                       \
    }

#define DXX_END_DSQL_TIMING_WITH_STATS(str,acc_type,cmp_key_cnt,prel)     \
    if (TimingsOn) {                                                      \
        _ftime(&Time1);                                                   \
        TimeDif =  (Time1.time - Time0.time) * 1000                       \
             + (Time1.millitm - Time0.millitm);                           \
        if (TRACE_TEST(pThrd, SL31)) {                                    \
           TraceSQLStmt(pThrd, str, TimeDif);                             \
           RDB_TraceTimings(RDB_TIMING_TRACE);                            \
        }                                                                 \
        if (pDXCB->BuildStats) {                                          \
            BuildSQLStatsStruct(pThrd, pDXCB,                             \
                                acc_type,                                 \
                                TimeDif,                                  \
                                cmp_key_cnt,                              \
                                prel,NULL);                               \
        }                                                                 \
        memset(&(pRDBThrdCB->Timings), '\0',                              \
               sizeof(pRDBThrdCB->Timings));                              \
    }


#define DXX_END_DSQL_TIMING_WITH_STATS1(pWA, str,acc_type,cmp_key_cnt,prel) \
    if (TimingsOn) {                                                      \
        _ftime(&Time1);                                                   \
        TimeDif =  (Time1.time - Time0.time) * 1000                       \
             + (Time1.millitm - Time0.millitm);                           \
        if (TRACE_TEST(pThrd, SL31)) {                                    \
           TraceSQLStmt(pThrd, str, TimeDif);                             \
           RDB_TraceTimings(RDB_TIMING_TRACE);                            \
        }                                                                 \
        if (pDXCB->BuildStats) {                                          \
            BuildSQLStatsStruct(pThrd, pDXCB,                             \
                                acc_type,                                 \
                                TimeDif,                                  \
                                cmp_key_cnt,                              \
                                prel,pWA);                                \
        }                                                                 \
        memset(&(pRDBThrdCB->Timings), '\0',                              \
               sizeof(pRDBThrdCB->Timings));                              \
    }


#define DX_LTRIM(pstr)                  \
{                                       \
    while (*(pstr) == L' ')  (pstr)++;  \
}

#define DX_RTRIM(pstr)                          \
{                                               \
    WCHAR *  pstr2;                             \
    pstr2 = pstr + wcslen(pstr) - 1;            \
    while (pstr2 >= pstr && *(pstr2) == L' ')   \
        *(pstr2) = L'\0';                       \
        (pstr2)--;                              \
}



/* The following structure is used for keeping a simple list of
   columns.  Each entry has a pointer to a column name string. The
   Column name is not necessarily in the string sink.                 */
typedef struct s_DXColList
{
    short          NumCols;         /* num of Col entries     */
    struct         s_DXColEntry
    {
        wchar_t *    pColName;
    }  Col[1];
} DX_COL_LIST;

/* this macro calculates the DX_COL_LIST size for n Col entries */
#define DX_COL_LIST_SIZE(n)    ((sizeof(DX_COL_LIST)) +       \
                            (((n)-1)*sizeof(struct s_DXColEntry)))


/* The following structure is used for keeping a list of extended
   columns column attributes.  For a given table, these entries match
   up with the COL_DEF entries in the Table Col List.                 */
typedef struct s_DXColAttrList
{
    short    NumCols;            /* num of Col entries     */
    struct   s_DXColAttrEntry
    {
        wchar_t * pUserType;        /* If a user defined data type, a
                                       ptr to a ZDlm string of the name. */
        long      DBMSLen;          /* DBMS Length for UDB/DB2/Informix  */
        short     ColNum;           /* CSS original Column position in the
                                      table. 1 based                    */
        PST_BIT8  Updatable:   1;  /* Col may be updated.               */
        PST_BIT8  Insertable:  1;  /* Col may be included on INSERT.    */
        PST_BIT8  DefaultData: 1;  /* Col has a default data clause     */
        // BJO 04/20/2009   IR 72718   DX_COL_ATTR_LIST   (add bAsChar)
        PST_BIT8  bAsChar:     1;  /* Col has character semantics       */
    }  Col[1];
} DX_COL_ATTR_LIST;

/* this macro calculates the DX_COL_ATTR_LIST size for n Col entries */
#define DX_COL_ATTR_LIST_SIZE(n)    ((sizeof(DX_COL_ATTR_LIST)) +     \
                            (((n)-1)*sizeof(struct s_DXColAttrEntry)))



/* This struct maintains info about an Index to be created for a table. */
typedef struct s_DXIndex
{
    struct s_DXIndex *  pNext; /* ptr to the next Index struct.
                                  This list is anchored in the
                                  corresponding DX_TABLE.          */
    void *      hIndex;        /* IXB Index handle                 */
    wchar_t     IndexName[DX_INDEX_NAME_LEN+1];  /* Index Name     */

    short       NumCols;        /* Num Cols in the following array */
    short       KeyColNum[16];  /* Index into pRowColList of the
                                    columns that make up index key */
} DX_INDEX;




/* This structure maintains info about the Primary Key for a table.
   If it is a joined view, there may be multiple Primary Keys used.
   One that is the concatenation of all the Base Table PKs, and
   potentially one for each Base Table PK.  Also, Extended Rels
   may have their own DX_PK if the Parent table expressions do not
   match any of the other PK Cols. */

typedef struct s_DXPK
{
    struct s_DXPK *   pNext;            /* ptr to the next PK struct.
                                           This list is anchored in the
                                           corresponding DX_TABLE. */

    struct s_DXTable *  pDXTable;       /* tbl containing the PK */

    short         PKType;               /* DX_PRIMARY - the PK for
                                               the table or view
                                           DX_EXTENDED - the PK cols for
                                               a joined view base tbl or
                                               an extended rel */
    DX_COL_LIST *  pPKSimpleColList;    /* List of columns that make up
                                           the PK.  NULL if there is no
                                           PK or this DX_PK is for an
                                           Extended Rel with exprs    */
    RDB_COL_LIST * pPKAreaColList;      /* ColList for the PK Area. These
                                           are the RHS values for the
                                           Par's SELECT and the Temp
                                           Tbl Key values.            */
    short         PKAreaLen;            /* PK Area length */
    COL_XR_TRANS_HANDLE  hParRecToPKArea;  /* Transformation from Par
                                           rec to PK Area.            */

    COL_XR_TOKEN *    p1stParLHSToken;  /* Ptr to 1st Par Tbl LHS Token
                                           to be used in building the
                                           Par Tbl SELECT WHERE PK = ?*/
    short   PKColMayBeNULL;             /* Flag indicating if any PK
                                           col may be NULL. T|F       */
    UCHAR   VarCharCompare;             /* Rule for building this PK's
                                           compare definition.
                                           RDB_PADDED_COMPARE or
                                           RDB_NOT_PADDED_COMPARE     */
    UCHAR   BinPadCharDiff;             /* TRUE or FALSE, for special
                                           PKs setup for rels when the
                                           Bin Pad char is different for
                                           the Par and Chl Tables.    */

    UCHAR   ForceTblAccess;      /* Directives to force how Extract will
                                    access rows via this PK:
                                        DX_FORCE_NONE   - Don't force
                                        DX_FORCE_SCAN   - Force Tbl Scan
                                        DX_FORCE_KEYLOOKUP - Force
                                              Select via PK columns   */
    short    ParCompoundKeyLookupCnt;  /* Number of key lookups to
                                       generate into the Where Clause
                                       for Parent Table Lookups.      */
    UCHAR   IndexAvailDetermined;       /* T|F, the following field has
                                           been set appropriately.    */
    UCHAR   IndexAvail;                 /* T|F, An index is avail to
                                           support lookup by this PK  */
    UCHAR   UniqueIndex;                /* T|F, A unique index exists
                                           for this PK                */
    UCHAR   IXMIndexDet;                /* T|F, IXM index availibility
                                           has been determined.       */
    SEL_KEY_LOOKUP_PARMS *  pKyLkParms; /* Parm struct used when
                                           accessing an IXM index for
                                           this PK.                   */

    /* Select statement using this PK */
    wchar_t *     pPKSelect;            /* SELECT ... WHERE PK = ?   */
    long          PKSelectLen;
    long          PKSelAllocLen;        /* MemAlloc Len of above SEL */
    float         SelectCost;           /* DB2 estimated cost of the
                                           above Select Stmt */
    UCHAR         CostDetermined;       /* Flag indicating if the above
                                           cost has been obtained */
    /* PK Temp Table used during extract */
    UCHAR         TempTableNeeded;      /* A Temp Table is needed
                                           for this PK, T|F */
    TTB_HANDLE       hPKTmpTbl;
    COL_CMPR_LIST *  pPKCmprList;       /* PK Compare definition */
    UCHAR *          pPKCmprCode;       /* Compiled Cmpr area.  Only
                                           present if Par and Chl
                                           table have diff Padded
                                           Compare Rules.          */
    UCHAR *          pPKArea;           /* PK record area */
    UCHAR *          pTmpTblRec;        /* Tmp Tbl IO Area */
    UCHAR *          pTmpTblRec2;       /* Aux Tmp Tbl IO Area */
    short            TmpTblRecLen;      /* Length of the above rec */

    /* Some counters, etc. used during extract */
    long          NumPKToProcess;       /* Runtime counter */
    short         ChildCntArrayOffset;  /*Offset of the start of Cntrs*/

    /* status fields for parallel processing */
    TDB_RID       PST_VOLATILE TDBRid;
    UCHAR         PST_VOLATILE bProcessed;
} DX_PK;


// DX_NAMEPIECE is a linked list of pieces that will be used to construct
// filenames for External File References
typedef struct sDXNamepiece
{
    struct sDXNamepiece * pNext;     // Next Entry
    UCHAR  Type;                     // Column or literal Indicator
                                     // (see defines where DX_TBL_NAMEPIECE is declared)
    WCHAR *pValue;                   // Pointer to the colname or literal
    RDB_COL_LIST_ENTRY *pColListEnt; // Col List entry for this piece's col
    long   LenValue;                 // Lth of pValue
} DX_NAMEPIECE;


// DX_ATTACH_DIRS is a linked list of directories where file attachments reside
typedef struct sDXAttachDirs
{
    struct sDXAttachDirs * pNext;    // Next Entry
    WCHAR                   * pDirName; // Pathname of directory
    UCHAR                     Used;     // T/F to indicate if any files were read from this dir
} DX_ATTACH_DIRS;

// DX_ATTACH_PARMS is a linked list of rules on how to handle file attachments
typedef struct sDXAttachParms
{
    struct sDXAttachParms     * pNext;   // Next Entry
    DX_ATTACH_DIRS      * pAttachDirs;   // List of directories to find attachments
    DX_NAMEPIECE        * pNamePieces;   // Pieces of filenames of attachments
    WCHAR               * pTriggerCol;   // Optional name of column that indicates
                                         //   whether row has an attachment
    UCHAR            StopWhenNotFound;   // T/F - Stop when file not found
    WCHAR            * pPseudoColName;   // Prefix for pseudo-column names
    UCHAR                  DeleteFile;   // T/F - Delete file during DEL process
    XFCOL            * pPseudoColFile;   // XF Col entry for Filename pseudocolumn
    XFCOL            * pPseudoColBLOB;   // XF Col entry for BLOB pseudocolumn
    XFCOL          * pPseudoColAttrib;   // XF Col entry for Attrib pseudocolumn
    PST_DOHANDLE                hFile;   // File handle for attachemnt
    UCHAR                    FileOpen;   // T/F - File attachment is open
    RDB_COL_LIST_ENTRY *pTrigColEntry;   // Col list entry for Trigger Column
    UCHAR *              pTrigColCmpr;   // Compiled compare for Trigger Column
} DX_ATTACH_PARMS;

/* This structure maintains info about the tables being extracted.
   Each cell has a linked list of these entries to define the tables
   in the Cell. */

typedef struct s_DXTable
{
    struct s_DXTable *   pNext;         /* ptr to the next tbl struct.
                                           This list is anchored in the
                                           DXCB. */

    struct s_DXTable *   pNextInCell;   /* ptr to the next tbl struct
                                           in the same cell.  This list
                                           is anchored in the DXCB. */
    void *  pCell;                      /* Ptr to the par cell struct */

    DX_INPUT_TABLE *    pInputTbl;      /* Ptr to the corr. TBL struct
                                           in the Input Table List    */
    XFTABLE *   pCopyXFTable;           /* If there is an Copy Extract
                                           file AND this table is on
                                           that XF, this points to
                                           the corresponding XF Table
                                           struct.                    */
    XFTABLE *   pInputXFTable;           /* If there is an Input Extract
                                           file, this points to
                                           the corresponding XF Table
                                           struct.                    */
    XFTABLE *   pCacheXFTable;           /* If there is an Input Extract file,
                                            this points to the special XFTABLE
                                            entry used by XFM_ConditionalValidateTable */


    wchar_t *   pWHEREClause;            /* Ptr to the AD's WHERE clause.
                                          NULL is none defined. */

    DX_ATTACH_PARMS *pAttachParms;       /* Ptr to first Attachment parms entry */

    /* Some extra clauses the user may have provided via commented
       hints ih the Where clause.  Currently tis is only valid
       for Oracle Start and ref tables.                            */
    long         fHintPresent;
    wchar_t *    pOrigWhereClause;
    wchar_t *    pDBMSHint;
    wchar_t *    pHintCorrName;
    wchar_t *    pHintFromClause;

    wchar_t * pDBAliasName;            /* Table's DBAlias Name   */
    wchar_t * pCreatorID;              /* Table's Creator        */
    wchar_t * pTableName;              /* Table's Name           */
    wchar_t * pCorrName;               /* Optional Correlation Name to
                                         be used with SQL for this Table*/
    wchar_t * pDB2MVSSynCreatorID;     /* If DB2MVSSyn, then this fld
                                         has the base Creator ID   */
    wchar_t * pDB2MVSSynTableName;     /* If DB2MVSSyn, then this fld
                                         has the base Table Name   */
    wchar_t * pInputCreatorID;         /* Input parm's Table Creator*/
    wchar_t * pInputTableName;         /* Input parm's Table Name   */

    RDB_CONN_HANDLE  hConn;              /* Table's Connection Handle */
    RDB_CONN_ATTR *  pConnAttr;          /* Connection Attrinutes     */
    PST_TTOKEN       hConsistFilter;     /* DSQL Connection consistency
                                            filter.  NULL if not
                                            needed.                   */
    UCHAR   fUseTmpLOBFiles;  /* Table must stage LOB data to temp
                                  files before Archive BefRow Action
                                  is called                           */
    UCHAR   RefTable;         /* Table is a ref table, T|F            */
    UCHAR   OrderTable;     /* Table is a reference table ot the
                               Start Table and the rows should be
                               be put to the XF in PK order.
                               If the table is the Start Table and
                               it involved in a cycle or has a non-PK
                               Point and Shoot List, rows will not
                               be in PK order.
                               Ignored for DX_FUNC_EXEC calls         */
    UCHAR   Type;                       /* Tbl type, RDB_TABLE |
                                           RDB_VIEW                   */
    UCHAR   DB2MVSSyn;                  /* T|F, DB2 MVS Synonym       */
    UCHAR   DBMSType;
    UCHAR   TableTraversed;             /* Tbl will be traversed during
                                           the extract, T|F. */
    UCHAR   AnyUDTs;                    /* One or more cols has a user
                                           defined Data Type, T|F     */

    UCHAR   UseDescribeForCost;       /* T|F, use SQL Describe to
                                         determine if an index is avail
                                         instead of fetching indexes. */

    UCHAR           IndexesFetched;     /* T|F Indexes have been fetched
                                           for this table.            */
    DX_COL_LIST *   pIndexColList;      /* List of cols that are the
                                           first column in an index on
                                           this table. NULL if no Idxs*/
    DX_INDEX *      pDXIndex;           /* Ptr to a linked list of
                                           Archive Indexes to be created
                                           for this table.              */
    XFTABLE *       pXFTable;           /* Ptr to corresponding output
                                           XF Table                     */

    void *  pInCommingPathList;         /* Ptr to a linked list of
                                           paths where this table is
                                           the destination. */
    void *  pOutGoingPathList;          /* Ptr to a linked list of
                                           paths where this table is
                                           the start. */

    long       RowLen;                  /* Data length of the row */
    short      NumCols;                 /* Num of cols in the list */
    XFCOL *    pColList;                /* Ptr to the col list */
    DX_PK *    pPKList;                 /* Ptr to the list of PKs for
                                           this table.  Joined views
                                           may have multiple. */
    DX_COL_LIST *  pSortColList;        /* Cols to sort the rows by on
                                           the Extract File           */

    long        TblId;                  /* Table ID in Extract File, nth
                                           table on the XFTABLE list, 1
                                           based.                     */

    /* Select statements used during the extract */
    WCHAR *       pSimpSelect;          /* SELECT with no Key lookup */
    long          SimpSelectLen;
    long          SimpSelAllocLen;      /* MemAlloc Len of above SEL */
    long          OffsetToWHERE;        /* Index into the statement to
                                           the space before the WHERE.
                                           A NULL may be moved here to
                                           lop off the WHERE clause.
                                           Valid even if no WHERE. */
    long          RowCount;             /* Number of rows in the table
                                           or view. 1 when RUNSTATS
                                           has not been run. */
    long          RowFailCount;         /* Number of rows we tried to
                                           read but couldn't because
                                           the BLOBs and CLOBs are too
                                           long.                      */
    long          NumDBMSXlateErrs;     /* Num of times a row failed
                                           because of a DBMS data
                                           translation error          */
    float         SelectCostwWHERE;     /* DB2 estimated cost of the
                                           above Select Stmt with the
                                           WHERE clause included */
    float         SelectCostwoWHERE;    /* DB2 estimated cost of the
                                           above Select Stmt with out
                                           the WHERE clause included */
    UCHAR         RunStatsAvail;        /* Have RUNSTATS been run for
                                           this table, T|F */
    UCHAR         CostDeterminedwWHERE; /* Flag indicating if the above
                                           cost has been obtained */
    UCHAR         CostDeterminedwoWHERE; /* Flag indicating if the above
                                           cost has been obtained */
    RDB_COL_LIST *      pTblColList;     /* Table's ColDef */
    RDB_MEM_HANDLE      hMemColList;     /* RBD Mem Handle of ColList */
    DX_COL_ATTR_LIST *  pColAttrList;    /* List of Extended Col Attrs*/
    RDB_CURSOR_HANDLE   hCursor;         /* Cursor Handle for this tbl*/

    void *              hSELEval;        /* SEL Component handle to
                                            evaluate a WHERE clause */

    UCHAR   SELCondAlwaysFALSE;          /* SEL Component determined
                                            that the WHERE clause
                                            always evaluates to FALSE */
    UCHAR   fTableHasLOBs;               /* T|F, table has LOB Columns */

    /* Some counters, etc. used during extract */
    short         AllRowsSelected;    /* No further processing is needed
                                         for this table since all rows
                                         have been selected. */
    long          NumRowsPut;         /* Num of rows put to the file */
    long          SelectRowLimit;
    long          SelEveryNthRow;     /* select all rows if this = 0 */
    long          NthRowCount;        /* Temp counter to keep track of
                                         how many rows were skipped
                                         since the last one was sel'd */

    /* Following flds define some attrs about the table's Connection  */
    UCHAR               StringDelim;
    UCHAR               DecPoint;
    UCHAR               BinPadChar;
    UCHAR               VarCharCompare;
    RDB_EXPR_SEMANTICS  ExprSemantics;

    /* The following are statistics about how the table was accesses
       and the elapsed time to process the table.                     */
    long    NumTblScans;      /* Num times tbl was processed via Scan */
    long    NumTblKeyLookups; /* Num times processed via Key Lookup   */
    float   ScanTime;         /* Elapsed time in secs to do Scans     */
    float   KeyLookupTime;    /* Elapsed time in sec to do Key Lookups*/

    /* Archive Fields */
    void *  hBefTblAction;
    UCHAR   fBefTblActionCalled;    /* TRUE|FALSE, action was called   */
    UCHAR   BefTblActionDBMSType;
    UCHAR   BefTblActErrControl;    /* DX_ACTION_ERR_STOP, ...         */
    RDB_CONN_HANDLE hBefTblActionConn;
    void *  hAftTblAction;
    UCHAR   AftTblActionDBMSType;
    UCHAR   AftTblActErrControl;    /* DX_ACTION_ERR_STOP, ...         */
    RDB_CONN_HANDLE hAftTblActionConn;
    void *  hBefRowAction;
    UCHAR   BefRowActionDBMSType;
    UCHAR   BefRowActErrControl;    /* DX_ACTION_ERR_STOP, ...         */
    RDB_CONN_HANDLE hBefRowActionConn;
    XFCOL *     pPseudocolRRN;       // if AFMF, current table RRN pseudo column

    UCHAR   fUsedTPT;               /* TRUE if accessed using Teradata Parallel Transporter */
} DX_TABLE;




/* This structure maintains a linked list of relationships involved in
   the extract.  */


typedef struct s_DXRel
{
    struct s_DXRel *   pNext;           /* ptr to the next tbl struct */

    DX_INPUT_REL *   pInputRel;         /* Ptr to the corresponding rel
                                           in the input List          */

    DX_TABLE *  pChlDXTable;               /* Ptr to the Child table's
                                              DX Table entry */
    DX_TABLE *  pParDXTable;               /* Ptr to the Parent table's
                                              DX Table entry */

    /* These next three names identify the Child Table or View as
       identified in the DX_INPUT_TABLE List                          */
    wchar_t *     pChlDBAliasName;
    wchar_t *     pChlCreatorID;
    wchar_t *     pChlTableName;
    /* These next three names identify the Parent Table or View as
       identified in the DX_INPUT_TABLE List                          */
    wchar_t *     pParDBAliasName;
    wchar_t *     pParCreatorID;
    wchar_t *     pParTableName;
    /* These next four names identify the Relationship Name.  If the
       above Child names identify a table, then the following
       DBAliasName, CreatorID, and TableName are the same as
       pChlDBAliasName, pChlCreatorID, and pChlTableName.  However, if
       the above Child names identify a view, then the following names
       are the true relationship name as defined to the view's base
       table.                                                         */
    wchar_t *     pRelDBAliasName;
    wchar_t *     pRelCreatorID;
    wchar_t *     pRelTableName;
    wchar_t *     pRelName;
    UCHAR       Type;    /* Relationship Type,  RDB_DBMS, RDB_PST or
                            RDB_GENERIC.                              */

    UCHAR   RelTraversed;               /* Rel will be traversed during
                                           the extract, T|F.   May be
                                           set to FALSE if the REL expr
                                           can never be valid.        */
    UCHAR   AlwaysREJECTED;             /* The relationship can never
                                           transform without REJECT   */
    UCHAR   AlwaysGetChildren;          /* Follow this rel in the par to
                                           child direction if the parent
                                           row was selected via a parent
                                           path. T|F */
    UCHAR   AlwaysGetParents;           /* Follow this rel in the child
                                           to par direction. T|F */

    UCHAR   IndexAvailDetermined;       /* T|F, the following field has
                                           been set appropriately.    */
    UCHAR   IndexAvail;                 /* T|F, An index is avail to
                                           support lookup of Chl Tbl
                                           via FK                     */
    UCHAR   IXMIndexDet;                /* T|F, IXM index availibility
                                           has been determined.       */
    SEL_KEY_LOOKUP_PARMS *  pKyLkParms; /* Parm struct used when
                                           accessing an IXM index for
                                           this FK.                   */

    UCHAR   ForceParTblAccess;   /* Directives to force how Extract will
                                    access Parent rows related to
                                    Extracted Child rows for this rel:
                                        DX_FORCE_NONE   - Don't force
                                        DX_FORCE_SCAN   - Force Tbl Scan
                                        DX_FORCE_KEYLOOKUP - Force
                                              Select via PK columns   */

    UCHAR   ForceChlTblAccess;   /* Directives to force how Extract will
                                    access Child rows related to
                                    Extracted Parent rows for this rel:
                                        DX_FORCE_NONE   - Don't force
                                        DX_FORCE_SCAN   - Force Tbl Scan
                                        DX_FORCE_KEYLOOKUP - Force
                                              Select via FK columns   */

    RDB_REL_DATA       RelData;         /* REL Definition             */
    RDB_MEM_HANDLE     hMemRelData;     /* RBD Mem Handle of RelData  */

    /* Some stuff that is needed during extract and other processes,
       but not kept on the extract file.  */
    DX_PK *       pPKDef;            /* Ptr to the parent table's DX_PK
                                        struct that defines the PK for
                                        this relationship. */
    RDB_COL_LIST *  pFKAreaColList;  /* ColList for the FK Area.  These
                                        are the RHS values for the Chl
                                        Tbl's SELECT.                 */
    COL_XR_TOKEN *  p1stChlLHSToken; /* Ptr to 1st Chl Tbl LHS Token to
                                        be used in building the Chl Tbl
                                        SELECT WHERE FK = ?           */
    wchar_t *     pFKSelect;        /* The SELECT statement
                                       to get rows via FK
                                       values from the child tbl. */
    long          FKSelectLen;
    long          FKSelAllocLen;    /* MemAlloc Len of above SEL */
    float         SelectCost;       /* DB2 estimated cost of the above
                                       Select Stmt */
    long          NumFKToProcess;   /* Num of FK's in the temp table
                                       that need to be processed */
    UCHAR         CostDetermined;   /* Flag indicating if the above
                                       cost has been obtained */

    /* Some control info used during extract to specify where the
       FKProcNeeded, FKProcDone, and FKInsThisPass flags are within
       the PathFlags array in the PK Temp Table.  These flags are used
       for Child Paths. */
    short          FlagByteIndex;      /* Index to be used when using
                                          TmpTblRec.PathFlags */
    UCHAR          FKProcNeededMask;   /* Mask with one bit on that corr
                                          to the FKProcNeeded flag */
    UCHAR          FKProcDoneMask;     /* Mask with one bit on that corr
                                          to the FKProcDone flag */
    UCHAR          FKInsThisPassMask;  /* Mask with one bit on that corr
                                          to the InsertThisPass flag */

    /* The following fields are used when following a parent path.
       In other words, when a child row is processed, the FK is
       extracted and put into the parent table's PK temp table and
       processed later to find the corresponding parents. */
    COL_XR_TRANS_HANDLE hChlRecToPKArea; /* Transform from the Chl Tbl
                                        Rec Cols to the Par Table TmpTbl
                                        PK Area                       */

    /* The following fields are used when following a child path.     */
    short         FKAreaLen;         /* FK Area length                */
    UCHAR *       pFKArea;           /* Ptr to the FK Area            */
    short         FKColMayBeNULL;    /* Flag indicating if any FK
                                        col may be NULL. T|F          */
    COL_XR_TRANS_HANDLE hPKAreaToFKArea; /* Transformat from the Tmp Tbl
                                        PKArea Cols to the RHS operands
                                        needed for the Chl Tbl SELECT */
   short   ChildLimit;               /* Limit num of children to get
                                        -1 : Select no children
                                         0 : No Limit
                                         n : Only n children per par  */
                                         /* DF_UNLIMITED_CHILDREN     */
   short   ChildCntOff;              /* Only set if ChildLimit > 0    */
                                     /* Contains offset from start of */
                                     /* Counter Array                 */

    short    ParCompoundKeyLookupCnt;  /* Number of key lookups to
                                       generate into the Where Clause
                                       for Parent Table Lookups.      */
    short    DepCompoundKeyLookupCnt;  /* Number of key lookups to
                                       generate into the Where Clause
                                       for Dependent Table Lookups.   */
    /* status fields for parallel processing */
    TDB_RID       PST_VOLATILE TDBRid;
    UCHAR         PST_VOLATILE bProcessed;

    //RT: Ignore Empty Key
    DX_REL_IGN_OPTIONS  *IgnoreOptions;  /* Bit masked variable
                                             bit 1 -> Null
                                             bit 2 -> Blank
                                             bit 3 -> Zero Length
                                             bit 4 -> Number
                                          */
    //RT: Ignore Empty Key

} DX_REL;




/* This structure maintains info about the paths followed during an
   extract.  There is a linked list of all used paths anchored in
   the DXCB.  Also, each table struct has a linked list of InComming
   and Outgoing paths from that table. */

typedef struct s_DXPath
{
    struct s_DXPath *  pNext;          /* ptr to the next path.  This
                                          list is a list of all paths
                                          followed in the traversal. */
    struct s_DXPath *  pNextInComming; /* Ptr to the path with the same
                                          table as the destination of
                                          the path. */
    struct s_DXPath *  pNextOutGoing;  /* Ptr to the path with the same
                                          table as the start of the
                                          path. */

    DX_REL *       pRel;               /* Ptr to the corresponding rel
                                          structure */

    DX_TABLE *     pParTable;           /* Ptr to the parent table */

    DX_TABLE *     pChlTable;           /* Ptr to the child table */

    UCHAR          PathType;          /* DX_CHILD_PATH or DX_PAR_PATH */

    UCHAR          CyclePath;           /* Path is a cycle path, T|F. */

    UCHAR          AlwaysGetChildren;   /* Follow this rel in the par to
                                           child direction if the parent
                                           row was selected via a parent
                                           path. T|F */
    UCHAR          Processed;           /* A flag used during
                                           BuildTheCells processing */
} DX_PATH;



/* This structure maintains info about the cells visited during an
   extracted.  */

typedef struct s_DXCell
{
    struct s_DXCell *  pNext;          /* ptr to the next cell.  This
                                          list is anchored in the
                                          Strategy struct. */

    DX_TABLE *     pTable;             /* Ptr to the list of tables in
                                          the cell. */

    UCHAR          CycleCell;          /* This is a cycle cell, T|F. */
    UCHAR          DeadEnd;            /* There are no outgoing paths
                                          from this cell, T|F */

} DX_CELL;






/* A linked list linked list of arrays is used in two cases.  First,
   on request, AllPathTraversal function will record each path taken
   at each step away from the start.  This is requested during the
   final cell traversal.  From the previous list, a new set of lists
   are built, one from each "step" that defines the paths that are
   followed from that cell. */

typedef struct s_DXCellExitPaths
{
    UCHAR                        DumpID[8]; /* "DXCLEXXX" */
    struct s_DXCellExitPaths *   pNext;     /* Ptr to the next list   */
    long                         NumPaths;  /* num of Path occurances */
    struct   s_ExitPath
    {
       DX_PATH *    pDXPath;        /* Ptr to the path at the start of
                                       this path. */
       DX_CELL *    pStartCell;     /* Ptr to the cell at the start of
                                       this path. */
       DX_CELL *    pEndCell;       /* Ptr to the cell at the end of
                                       this path. */
    }  ExitPath[1];             /* Array of Exit Path entries */

} DX_CELL_EXIT_PATHS;

/* this macro calculates the DX_CELL_EXIT_PATHS size for n entries */
#define DX_CELL_EXIT_PATHS_SIZE(n)    ((sizeof(DX_CELL_EXIT_PATHS)) +  \
                            (((n)-1)*sizeof(struct s_ExitPath)))





/* A linked list of these structures is built to define the order the
   cells must be visited to perform the extract. */

typedef struct s_DXStep
{
    UCHAR          DumpID[8];       /* "DXSTEPXX" */
    short          NumSteps;        /* num of Step occurances */
    struct   s_StepEntry
    {
        DX_CELL *    pCell;         /* Ptr to the cell to be processed
                                       this step. */
        DX_CELL_EXIT_PATHS *
                     pExitPaths;    /* The list of paths that are
                                       followed after this cell is
                                       processed. */
        short        LevelNum;      /* This step is logically n path
                                       steps from the start table. */
        UCHAR        CellRefLater;  /* Flag indicating if this cell is
                                       ref'd in a later step, T|F */
    }  Step[1];                   /* Array of path entries */

} DX_STEP_LIST;

/* this macro calculates the DX_STEP_LIST size for n step entries */
#define DX_STEP_LIST_SIZE(n)    ((sizeof(DX_STEP_LIST)) +       \
                            (((n)-1)*sizeof(struct s_StepEntry)))



/* The following record is built during extract to contain host
   variable values that may be used as input to Archive Action
   Statements.                                                   */
/* For UNIX, these fields need to be properly aligned.  Although
   unnecessary, all fields will start on a 4 byte boundry.       */
typedef struct s_DXArchHostVars
{
    long    ArchiveID;         /* Sequentially assigned Archive File ID.   */
    PST_WCHAR_LEN ArchiveFileNameLen;   /* Len of following field          */
    wchar_t ArchiveFileName[250 + 1 + 3];   /* RDB_DSN_NAME_LEN */
                               /* Archive File Name                        */
    long    ActionType;        /* DML_ACTION_... type for action           */
    short   ActionTypeTextLen;   /* Len of following field                 */
    UCHAR   ActionTypeText[20+1+1];  /* Action Type Text                   */

    PST_WCHAR_LEN IndexFileNameLen;  /* Len of following field             */
    wchar_t IndexFileName[250+1+3];   /* RDB_DSN_NAME_LEN */
                               /* Archive File's Index File                */
    PST_WCHAR_LEN GroupNameLen;      /* Len of following field             */
    wchar_t GroupName[8+1+1];  /* DX_GROUP_NAME_LEN                        */
                             /* Group the Archive File is a member of      */
    PST_WCHAR_LEN UserIDLen;       /* Len of following field               */
    wchar_t UserID[30+1+3];  /* RDB_USERID_LEN                             */
                             /* Workstation ID of the user who executed the
                                Archive                                    */
    PST_WCHAR_LEN ServerNameLen;      /* Len of following field            */
    //KSB: 07/24/09 :Support for long Computer Name
    //wchar_t ServerName[15+1+2]; /* DX_SERVER_NAME_LEN                      */
    wchar_t ServerName[OPTIM_MAX_COMPUTERNAME_LENGTH+1+2];
                           /* Server name used to produce the Archive File */
    PST_WCHAR_LEN MachineNameLen;      /* Len of following field           */
    //KSB: 07/24/09 :Support for long Computer Name
    //wchar_t MachineName[15+1+2]; /*  DX_SERVER_NAME_LEN */
	 wchar_t MachineName[OPTIM_MAX_COMPUTERNAME_LENGTH+1+2];
                           /* Machine name used to produce the Archive File*/
    PST_WCHAR_LEN DescriptionLen;        /* Len of following field         */
    wchar_t Description[128+1+1];  /* DX_LONG_DESC_LEN                     */
                            /* User descriprion of Archive File            */
    union uDXStrtDT
    {
        RDB_C_TIMESTAMP    DB2;
        RDB_SYB_DT         SybSqlSrvr;
        RDB_ORA_DT         Ora;
        RDB_INFX_DT        Infx;
        RDB_TERA_TIMESTAMP Tera;       /* ShilpaJ 05/25/2011 Added for Teradata Support */
    } StartDT;

    union uDXCurrDT
    {
        RDB_C_TIMESTAMP    DB2;
        RDB_SYB_DT         SybSqlSrvr;
        RDB_ORA_DT         Ora;
        RDB_INFX_DT        Infx;
        RDB_TERA_TIMESTAMP Tera;       /* ShilpaJ 05/25/2011 Added for Teradata Support */
    } CurrDT;

    short   DateCharLen;         /* Len of the following field           */
    UCHAR   DateChar[24+1+1];    /* Current date/time in char format     */

    long    TotRowsExtracted;    /* Total rows extracted for all tbls    */
    wchar_t Guid[38];            /* Archive File GUID                    */
    UCHAR   Slack1[2];           /* Filler                               */
    long    ExtractStatus;       /* The current state of the Process:
                                     DX_STATUS_SUCCESS
                                     DX_STATUS_CANCEL
                                     DX_STATUS_FATAL_DBMS_ERROR
                                     DX_STATUS_ACTION_ABORT
                                     DX_STATUS_FATAL_ERROR              */

    /* The following fields are for Table related Action Statements     */
    PST_WCHAR_LEN TblDBAliasNameLen;     /* Len of following field       */
    wchar_t TblDBAliasName[12+1+1];      /* RDB_DBALIAS_LEN              */
    PST_WCHAR_LEN   TblCreatorIDLen;     /* Len of following field       */
    wchar_t TblCreatorID[64+1+1];        /* RDB_CID_LEN                  */
    PST_WCHAR_LEN   TblNameLen;          /* Len of following field       */
    wchar_t TblName[64+1+1];             /* RDB_TBL_NAME_LEN             */
    /* KSB - 03/17/2011 - RTC-22144
       Uplift the Maximum Row limit supported in Optim */
    unsigned long long    TblRowsExtracted; /* Total rows extracted for this tbl    */
} DX_ARCH_HOST_VARS;


/* This structure is used to maintain a pool of connections to all DXTABLEs;
All of then are copies of handle in DXTABLE and must be closed after all
tables are processed */
typedef struct sDXConnPool
{
    short            MaxhConn; /* max no of DBMS handles in pool,
                                  unused entryes are zeroed          */
    struct{
        wchar_t         *pDBAliasName; /* DB Alias Name, see DXTABLE */
        RDB_CONN_HANDLE  hConn; /* Corresponding connection,
           a duplicate connection of the codresponding one in DXTABLE*/
        UCHAR            fInUse;
    } *pDbmsConn;
} DX_CONN_POOL;


/* This structure is the anchor for the extract strategy.  It is result
   during the DetermineExtractStrategy function. */

typedef struct s_DXCB
{
    UCHAR   Func;           /* DX_FUNC_SHOW_STEPS, DX_FUNC_EXEC, ...  */
    UCHAR   ValidateSQL;    /* TRUE | FALSE
                               For DX_FUNC_VALIDATE or DX_FUNC_COMPILE
                               calls, Prepare the SQL Select statements
                               as part of the validation process.
                               If a DX_FUNC_COMPILE_AND_EXEC or
                               DX_FUNC_EXEC call, prepare all the SQL
                               before starting the extract process.   */
    UCHAR   fValidateIndexes; /* T|F Func was VALIDATE_INDEXES        */
    UCHAR   BuildXFPKIndex; /* T | F.  Build a PK Index on the Extract
                               File.  SyncPoint only.                 */
    UCHAR   XtractFromXF;  /* TRUE | FALSE, extract source is an
                              file.                                   */
    UCHAR   UseInputPKDef;  /* If TRUE, then Catalog Services is NOT
                               called to obtain PK information about
                               tables.  Instead the KeySeqNum fields
                               in the DX_INPUT_COL structs are used
                               to determine the Table's PK columns.   */
    UCHAR   CountDBMSXlateErrs;  /* TRUE | FALSE, count the number of
                                    times for each table a row fails
                                    because of DBMS data translation
                                    errors                            */
    MEM_HEAP_HANDLE   hHeap;             /* Heap for all DX structs   */
    TDB_HANDLE        hTDB;              /* Temp DB Handle            */
    void *            hXFB;              /* Xtract File Builder Handle*/
    void *            hXFM;              /* Input XF XFM Handle       */
    SEL_HANDLE        hSELEnv;           /* Sel Environment handle. Used
                                            with ExtractFromXF and
                                            there is an index file    */
    SEL_HANDLE        hSELConn;          /* Sel Connection handle.  Used
                                            with ExtractFromXF and
                                            there is an index file    */
    long              XFMBlkSize;        /* Input XF BlkSize          */
    short             BlockDDLOnInputXF; /* TRUE|FALSE, there are
                                            DDL Plain blocks on the
                                            input XF                  */
    UCHAR             DDLEndian;         /* State of Input XF DDL blocks
                                             XFDDL_BIG_ENDIAN or
                                             XFDDL_LITTLE_ENDIAN      */
    UCHAR             DDLFmtWChar;       /* Size of Input XF DDL char */
    short             MvsDDLOnInputXF;   /* TRUE|FALSE, there are MVS
                                            DDL blocks on the input XF*/
    void *            hIXB;             /* IXB Index handle           */
    long              ArchiveID;         /* Archive File ID           */

    DX_INPUT_TABLE *  pInputTableList;   /* Input list of tbls/views  */

    DX_INPUT_REL *    pInputRelList;     /* Input list of rels        */

    /* KSB - 03/17/2011 - RTC-22144
       Uplift the Maximum Row limit supported in Optim */
    unsigned long long MaxRowsToPut;
                                         /* Maximum num of rows on the
                                            extract file */
    unsigned long long TotalRowsPut;     /* Current num of rows on the
                                            extract file */
    long        SaveRowsPut;             /* Save count for Stats
                                            processing */

    short       NumTbls;                 /* Num of tbls in the list */
    short       NumRels;                 /* Num of rels in the list */
    short       NumCells;                /* Num of cells in the list */
    long        NumPaths;                /* Num of Paths in the list */

    UCHAR *     pCopiedCompiledObj;      /* For Exec Only calls, the
                                            copied Extract Obj Area  */
    long        CopiedObjLen;            /* Length of above          */

    DX_TABLE *  pAllTblList;            /* Ptr to the 1st entry in the
                                           full table list */
    DX_REL *    pAllRelList;            /* Ptr to the 1st entry in the
                                           full Relationship List */
    DX_PATH *   pAllPathList;           /* Ptr to the 1st entry in the
                                           full path List */
    DX_CELL *   pAllCellList;           /* Ptr to the 1st entry in the
                                           full cell List */
    DX_STEP_LIST *  pStepList;          /* Ptr to the Processing Step
                                           List */
    DX_TABLE *  pStartTable;            /* Ptr to the start table     */
    XFTABLE *   pXFTableList;           /* XF List for XFB Open       */
    XFTABLE *   pInputXFTableList;      /* If an Extract of an XF, the
                                           XF List for XFM Open       */
    XFREL *     p1stInputXFRel;         /* If an Extract of an XF, the
                                           XFRels from the input XF   */
    UCHAR *     pFetchArea;             /* IO area used for Extract   */
    long        FetchAreaLen;           /* Len of the above area      */
    UCHAR *     pFullRowArea;           // Area containing row data
                                        // with pseudocolumn data appended
    long        FullRowAreaLen;         // Len of the above area
    wchar_t *   pStmtWorkArea;          /* Work area for blding Stmts */
    long        StmtWorkAreaLen;        /* Len of above area          */


    DX_ARC_COLLECTION * pFirstArcCollection; /* List of Archive Collection
                                                names to which the archive
                                                will be added               */

    UCHAR       MultiByteRTDisp;        // Round-trip error handling flag

    /* Pick List Control Info */
    UCHAR       PickListExistsFlag;    /* Flag indicating if a pick list
                                          was passed as input, T|F */
    short       UnionWithSelCrit;      /* Start with P&S List unioned
                                          with rows where SelCrit */

    void *      pProdAnchor;             /* Passed on all calls to
                                            product exits             */
                                         /* Exit for progress notify  */
    /* KSB - 03/17/2011 - RTC-22144
       Uplift the Maximum Row limit supported in Optim */
    short     (*pNotifyExit) (PST_THANDLE, void *, WCHAR *,
                              short, long, unsigned long long);
    DX_TABLE *  pNotifyTable;            /* Table to be referenced on
                                            the next Notify           */
    short     NotifyType;              /* Code for the next notify
                                          DX_PROCESSING IN_PROGRESS,
                                          DONE, or MAYBE_DONE       */
    long      NotifyLimit;             /* Increment when the User
                                          Progress Notify is done.  */
    long      NotifyTimeThreshold;   /* Elapsed time in millisecs between
                                        calls to the Notify Exit.     */
    PST_DODWORD LastNotifyTime;  /* Time last notify exit was called  */

    /* Error log exit - Called when XFB detects a duplicate row       */
    void      (*pErrorLogExit) (DX_LOG_CB *);

    /* The following field is polled during the Execution
       Phase to check if the calling product wants to abort the
       Extract.  A non-zero value in the byte pointed to by this
       field will abort the Extract.  If this field is NULL, no
       polling takes place.                                       */
    UCHAR *    pAbortSemaphore;

    short       InitStartTblPass;        /* Currently processing the
                                            first pass of the start
                                            table, TRUE | FALSE */
    short       fInACycle;                /* TRUE|FALSE, the current table
                                             is in an extract cycle */
    long        NumStartTable1stPassRows; /* Num rows selected from the
                                            start table on the its first
                                            pass */

    /* Some fields when using Aux Key Point & Shoot Lists */
    short           UsingNonPKColList;  /* TRUE | FALSE               */
    RDB_COL_LIST *  pInputPNSCols;      /* If UsingNonPKColList, the
                                           linked list of cols for
                                           the point and shoot value. */
    TTB_HANDLE      hAuxKeyTmpTbl;      /* Temp Table with aux keys   */
    long            NumAuxKeys;         /* Num of Aux Keys in the list*/
    DX_PATH *       pAuxKeyPath;        /* Data structs to simulate an
                                           incomming Child Path to the
                                           start table.               */
    RDB_COL_LIST *  pAuxKeyColList;

    UCHAR   RelNumCharConvMode;  /* Special flag for controlling how
                                    Relationship processing should
                                    handle Numeric to Character
                                    conversions.
                                      '0'
                                          Right justify num, zero fill
                                      '1'
                                          Right justify num, space fill
                                      '2'
                                          Left justify num, space fill*/

    /* Error control info */
    short       ErrCode;                  /* The error message number */
    wchar_t     ErrTok1[DX_ERRTOK_LEN+1]; /* Err msg substitution word*/
    wchar_t     ErrTok2[DX_ERRTOK_LEN+1]; /* Err msg substitution word*/
    wchar_t     ErrTok3[DX_ERRTOK_LEN+1]; /* Err msg substitution word*/
    wchar_t     ErrTok4[DX_ERRTOK_LEN+1]; /* Err msg substitution word*/

    /* the following array is used when selecting rows by Key value and
       one or more of the key cols has the NULL value.  This table
       keeps track of the select statements built so far for this table
       that have IS NULL for one or more key columns. */

    long           StmtUseCount;     /* Use TimStmp for stmt array */
    struct s_NULLStmt {
        unsigned long  StmtKey1;     /* mask of which cols are NULL   */
        unsigned long  StmtKey2;     /* mask of which cols are NULL   */
        void *         pStmtID;      /* if the statement is for a PK
                                        then DX_TABLE *.  If for a
                                        FK, the DX_PATH               */
        long           LastUseCount; /* TimeStamp of when last used   */
        wchar_t *      pStmt;        /* ptr to statement text         */
        long           StmtAllocLen; /* MemAlloc len of above stmt    */
        RDB_COL_LIST * pKeyColList;  /* ptr to corr. Key coldef       */
        long           ColListLen;   /* Len of above Col List         */
    } NULLStmt[NUM_NULL_STMTS];

    /* The following field is used to compute the CPU Time used by
       various steps in the extract */
    double   CPUTime;

    /* Some internal flags to force various execution actions */
    UCHAR   ForceKeyLookup;       /* T|F, Force Key lookup strategy   */
    UCHAR   ForceScan;            /* T|F, Force table scan strategy   */


    /* The following fields support n of m processing */
    short    NofMFlag;               /* T|F  N of M processing in use */
    wchar_t  NofMColumn[RDB_COL_NAME_LEN+1];
    DX_PK    *pNofMKey;              /* Ptr to N of M Key data        */
    long     NumRowsPerValue;        /* Max N Value                   */
    long     NumValues;              /* Max M Value                   */
    long     MaxNofMValues;          /* Max number of start table rows*/
    short    NofMOptimize;           /* T|F if close cursor optimize  */
    short    NofMCloseCursorFlag;    /* T|F if cursor should be closed*/

    /* The following two fields define attrs about the Directory DB */
    wchar_t  DirDBConnString[RDB_CONN_STR_LEN+1];
    UCHAR    DirDBMSType;

    /* Default names from the Access Definition to be copied to XFB */
    wchar_t  DefDBAliasName[RDB_DBALIAS_LEN+1];
    wchar_t  DefCreatorID[RDB_CID_LEN+1];

    /* The following fields help decide Table Scan versus Table Key
       Lookup access strategies.  The user tells us how often they
       want to have extract consider a Scan versus a Key Lookup to
       access a table.  The user can say never scan, rarely scan,
       or sometimes scan.  One of the following factors is choosen
       bas-ed on the user decision and the current situation; we have
       PKs to lookup and there is an appropriate index or FKs to lookup
       and an index is avail. If the following comparison is TRUE, the
       we use the Scan technique to fetch the rows:

            (NumKeysToLookup / NumRowsInTable) > Factor

      If a PK or FK is not backed by an index, we will do a scan unless
      the number of lookup keys divided by the number of rows is less
      than the corresponding ...WoIdx... factor.

      The Never factor should always be .99.  The comments suggest
      reasonable values for the other factors.                       */

    UCHAR   UserScanFactor;           /* DX_SCAN_NEVER,
                                         DX_SCAN_RARELY,
                                         DX_SCAN_SOMETIMES            */
    float   ParWIdxNeverFactor;       /* Access Par with avail indexes*/
    float   ParWIdxRarelyFactor;      /*   "     .7                   */
    float   ParWIdxSometimesFactor;   /*   "     .5                   */
    float   ParWoIdxNeverFactor;      /* Access Par wo avail indexes  */
    float   ParWoIdxRarelyFactor;     /*   "     .01                  */
    float   ParWoIdxSometimesFactor;  /*   "     .002                 */
    float   ChlWIdxNeverFactor;       /* Access Chl with avail indexes*/
    float   ChlWIdxRarelyFactor;      /*   "     .4                   */
    float   ChlWIdxSometimesFactor;   /*   "     .3                   */
    float   ChlWoIdxNeverFactor;      /* Access Chl wo avail indexes  */
    float   ChlWoIdxRarelyFactor;     /*   "      .005                */
    float   ChlWoIdxSometimesFactor;  /*   "      .001                */
    float   PKWIndexFactor;
    float   PKWoIndexFactor;
    float   FKWIndexFactor;
    float   FKWoIndexFactor;

    /* Statistics gathering fields */
    UCHAR        BuildStats;         /* TRUE|FALSE, build a linked list
                                        of extract statistics structs */
    DX_STATS_STEP *  p1stStatsStep;  /* First in a linked list of
                                        statistics structures         */
    DX_STATS_STEP *  pLastStatsStep; /* Last in a linked list of
                                        statistics structures         */
    MEM_HEAP_HANDLE  hRtnHeap;       /* Heap to use to alloc memory for
                                        statistics structures         */


    DX_OUTPUT_REL *  p1stRtnRel;     /* A returned linked list of
                                        structs that provide information
                                        about the relationships used     */

    /* Archive Action fields */
    void *   hBefAction;             /* Action handle for Before Extract */
    void *   hAftAction;             /* Action handle for After Extract  */
    UCHAR    BefActionDBMSType;
    UCHAR    AftActionDBMSType;
    UCHAR    BefExtrActErrControl;    /* DX_ACTION_ERR_STOP, ...          */
    UCHAR    AftExtrActErrControl;    /* DX_ACTION_ERR_STOP, ...          */
    RDB_CONN_HANDLE  hBefActionConn;
    RDB_CONN_HANDLE  hAftActionConn;

    RDB_COL_LIST *   DB2ActProcColList;  /* DB2 Sys HVs for Proc Act Stmts */
    RDB_COL_LIST *   DB2ActTblColList;   /* DB2 Sys HVs for Tbl Act Stmts */
    RDB_COL_LIST *   OraActProcColList;  /* Ora Sys HVs for Proc Act Stmts */
    RDB_COL_LIST *   OraActTblColList;   /* Ora Sys HVs for Tbl Act Stmts */
    RDB_COL_LIST *   SybActProcColList;  /* Syb Sys HVs for Proc Act Stmts */
    RDB_COL_LIST *   SybActTblColList;   /* Syb Sys HVs for Tbl Act Stmts */
    RDB_COL_LIST *   SqlSrvrActProcColList; /* SQLSrvr for Proc Act Stmts */
    RDB_COL_LIST *   SqlSrvrActTblColList;  /* SQLSrvr for Tbl Act Stmts */
    RDB_COL_LIST *   IfxActProcColList;  /* Ifx Sys HVs for Proc Act Stmts */
    RDB_COL_LIST *   IfxActTblColList;   /* Ifx Sys HVs for Tbl Act Stmts */
    RDB_COL_LIST *   TerActProcColList;  /* DB2 Sys HVs for Proc Act Stmts */
    RDB_COL_LIST *   TerActTblColList;   /* DB2 Sys HVs for Tbl Act Stmts */

    RDB_C_TIMESTAMP  StartTS;
    RDB_C_TIMESTAMP  CurrTS;

    DX_ARCH_HOST_VARS *   pActSysHVRec;

    /* Save areas for Action RDB Errors */
    struct s_DBMS_GenericErr    ActionRDBGenericErr;
    union  u_DBMS_SpecificErr   ActionRDBSpecificErr;

    //fields shared between parallel threads, and associated lock
    PST_TTOKEN                    LockTok;
    long                          MaxConn;
    DX_CONN_POOL                  ConnPool;

    // Fields for security
    wchar_t * pFADSName;             // File Access Data Structure name
    SEC_Token SecToken;              // XFM security token

    // Fields for file attachments
    short     NumFileNameChars;      // Max # chars in file attachment path/filename
    WCHAR *   pFileNameWorkArea1;    // Work area for constructing file attachment path/filename
    WCHAR *   pFileNameWorkArea2;    // Work area for constructing file attachment path/filename
    USHORT    FileAttribAreaSize;    // Size of Area for storing file attributes
    FILEATTR_BLOB *  pFileAttribArea;// Area for obtaining file attributes
    USHORT    FmtFileAttribAreaSize; // Size of area for storing formatted file attributes
    FILE_DATA * pFmtFileAttribArea;  // Area for formatting file attributes
    long        ExtractMaxAllocation;//DSQ Fetch Buffer size
    BOOL        FileComp;            // Is file compressed
    UCHAR       AddPseudcolRRN;      // Should a pseudocolumn be added for AFMF
    UCHAR       NoRICycleRevisit;    // If TRUE, do not revisit tables for RI cycle.
	// KSB - 06/14/2010 - Begin - Added for inline Active Compression
	short           sActCompDefThreshold;          // Active Compression Default Threshold.
	DXX_ACTIVE_COMPRESS *pFirstTableActCompBlk;    // Pointer to first table active compression blk.
	UCHAR		ActiveCompression;				   // Is Active Compression Enabled.
	XF_ACTCOMP  *pCommonActCompBlk;				   // Common Active Compression block.
	// KSB - 06/14/2010 - End- Added for inline Active Compression
} DXCB;


typedef struct s_DXParallelWorkArea
{
    //the following two connection handles must be used
    //in place of those in DX_TABLE
    RDB_CONN_HANDLE hConn;         /* Data access connection */
    RDB_CONN_HANDLE hBefRowActionConn; /* RowAction connection */
    PST_DOHANDLE    hTmpLOBFile;   /* for LOB extraction */
    UCHAR           fParallel;     /* TRUE/FALSE parallel mode indicator */
    short           ErrCode;
    PST_VOLATILE UCHAR *pAbortSemaphore; /* Points to same area as the DXCB one;
                                        Some macros use multiple fields of
                                        the same struct, so we duplicated
                                        this field only for this purpose */
    wchar_t         ErrTok1[DX_ERRTOK_LEN+1]; /* Err msg substitution word*/
    wchar_t         ErrTok2[DX_ERRTOK_LEN+1]; /* Err msg substitution word*/
    wchar_t         ErrTok3[DX_ERRTOK_LEN+1]; /* Err msg substitution word*/
    /* Save areas for Action RDB Errors */
    struct s_DBMS_GenericErr    ActionRDBGenericErr;
    union  u_DBMS_SpecificErr   ActionRDBSpecificErr;
    UCHAR          *pTmpTblRec;
    UCHAR          *pTmpTblRec2;
    UCHAR          *pFetchArea;
    long            hConsistFilter;
    UCHAR           bTraceKeys;
    UCHAR           bTraceBegEnd;
    UCHAR           bAbortSemaphore;
    short           RetCode;
    void*           pStructID;
    PST_DODWORD     CntRowsPut;
} DX_WORK_AREA;

/* A list of relationship cols from the parent or child table
   that are not involved in a substring or concatenation
   expression.  An index on these cols would be very nice. */
typedef struct s_DXPathCols
{
    short       NumCols;
    wchar_t *   pColName[128];

 } DX_PATH_COLS;



/* This structure is used during an All Paths Traversal among cells.
   When a path going into a cycle cell is followed, this list defines
   the list of outgoing cells that are followed from that cell. */

typedef struct s_APExitList
{
    short              NumExits;        /* num of Exit Path occurances */
    struct   s_APExitEntry
    {
        void *     pAPPathEntry;  /* Arbitrary rel ID, usually a ptr of
                                     some other structure. The two paths
                                     entries for the same relationship
                                     must have the same RelID */
    }  Exit[1];                   /* Array of path entries */

} DX_AP_EXIT_LIST;

/* this macro calculates the DX_AP_EXIT_LIST size for n exit entries */
#define DX_EXIT_LIST_SIZE(n)    ((sizeof(DX_AP_EXIT_LIST)) +       \
                            (((n)-1)*sizeof(struct s_APExitEntry)))








/* This structure is used by the AllPathsTraversal Function to define
   the input list of Paths */

typedef struct s_APPathList
{
    long              NumPaths;        /* num of Path occurances */
    struct   s_APPathEntry
    {
        void *     RelID;         /* Arbitrary rel ID, usually a ptr of
                                     some other structure. The two paths
                                     entries for the same relationship
                                     must have the same RelID */
        DX_PATH *  pDXPath;       /* for the final cell traversal, this
                                     points to the DX path for this AP
                                     path entry. */
        short      StartNode;     /* Subscript number in the Node List
                                     for the Node at the beginning of
                                     the path. */
        short      EndNode;       /* Subscript number in the Node List
                                     for the Node at the end of the.
                                     path. */
        struct s_APPathEntry *
                   pNextPath;     /* Ptr to the next Path entry in a
                                     linked list of paths that have
                                     the same node as the start of the
                                     path. */

        DX_AP_EXIT_LIST *
                   pExitList;     /* When entering a cycle node, this
                                     list defines the exit paths. */

        short      SeqNum;        /* The position of the path in the
                                     list. */
        short      LastStepNum;   /* The last step number that this path
                                     was traversed. This is set by the
                                     AllPathsTraversal function. */
        UCHAR      PathType;      /* DX_CHILD_PATH or DX_PAR_PATH */
        UCHAR      AlwaysGetChildren;  /* Follow this child path if
                                     the parent row was selected via
                                     a parent path. */
        UCHAR      StartPath;     /* This path is a path to start the
                                     traversal, T|F */
        UCHAR      PathTraversed; /* Path was traversed, T|F.  This is
                                     set by the AllPathsTraversal
                                     function. */
        UCHAR      CyclePath;     /* Is the path a cycle path?
                                     DX_UNKNOWN, DX_CYCLE_PATH,
                                     or DX_NOT_CYCLE_PATH,
                                     If AllPathsTraversal is about to
                                     follow this path and CyclePath =
                                     DX_NOT_CYCLE_PATH, the path is not
                                     followed. */

    }  Path[1];                   /* Array of path entries */

} DX_AP_PATH_LIST;

/* this macro calculates the DX_AP_PATH_LIST size for n path entries */
#define DX_PATH_LIST_SIZE(n)    ((sizeof(DX_AP_PATH_LIST)) +       \
                            (((n)-1)*sizeof(struct s_APPathEntry)))



/* This structure is used by the AllPathsTraversal Function to define
   the input list of Nodes */

typedef struct s_APNodeList
{
    UCHAR        DumpID[8];       /* "NODELIST" */
    short        NumNodes;        /* num of Node occurances */
    struct       s_APNodeEntry
    {
        void *     NodeID;        /* Arbitrary Node ID, usually a ptr of
                                     some other structure.  This is not
                                     used by the AllPaths Algorithm, but
                                     exists for the convience of the
                                     caller */
        struct s_APPathEntry *
                   p1stPath;      /* Ptr to the first Path entry in a
                                     linked list of paths where this
                                     node is the start of the path */

        UCHAR      NodeTraversed; /* Node was traversed, T|F.  This is
                                     set by the AllPaths Traversal
                                     function. */
        UCHAR      CycleNode;     /* this node is a cycle node.  When
                                     entered, the incomming path's
                                     exit list is used to determine the
                                     outgoing paths to follow. */
        UCHAR      resv[6];       /* Filler to make entry 16 bytes long.
                                     This makes for fast subscripting */

    } Node[1];

} DX_AP_NODE_LIST;

/* this macro calculates the DX_AP_NODE_LIST size for n node entries */
#define DX_NODE_LIST_SIZE(n)    ((sizeof(DX_AP_NODE_LIST)) +       \
                            (((n)-1)*sizeof(struct s_APNodeEntry)))


/* This structure defines the data area for the Temp Tables used during
   an extract.  There is at most one Temp Table per extract table.  The
   key for the temp table is the table's PK.  The following is the data
   portion of the Temp Table record. */
typedef struct s_DXTmpTblData
{
    /* These are the primary flags for the table record.  They include:
         DXTT_ROW_SELECTED
         DXTT_SEL_VIA_CHL_PATH
         DXTT_PROC_NEEDED
         DXTT_PROC_DONE
         DXTT_INSERT_THIS_PASS */
    UCHAR    TblFlags;

    /* Each byte in this array contains two bit flags for
       each of the table's outgoing paths.  They indicate if
       FK processing is needed and whether PK processing is
       done for this parent row's PK.  Which byte and bit flags
       within the byte is defined in the corr. DXPath entry. */
    UCHAR    PathFlags[1];

} DX_TMP_TBL_DATA;

/* this macro calcs the DX_TMP_TBL_DATA size for n path flag bytes */
#define DX_TMP_TBL_DATA_SIZE(n)   (sizeof(DX_TMP_TBL_DATA) + (n) - 1)



/* Bit flags for the TblFlags field */
/* Row for this PK is on the extract file */
#define   DXTT_ROW_PUT             0x80
/* Row for this PK was selected.  It may not be on the file because of
   Table limits or randomness selection */
#define   DXTT_ROW_SELECTED        0x40
/* Row for this PK has been selected via a child path */
#define   DXTT_SEL_VIA_CHL_PATH    0x20
/* A row with this PK needs to be selected */
#define   DXTT_PROC_NEEDED         0x10
/* Extract has already attempted to select a row with this PK value */
#define   DXTT_PROC_DONE           0x08
/* This temp table row was inserted on the current pass of the table */
#define   DXTT_INSERT_THIS_PASS    0x04
/* This is an original PK List entry */
#define   DXTT_PKLIST_ENTRY        0x02




/*--------------------------------------------------------------------*
 *  Data Extract Function Prototypes                                  *
 *--------------------------------------------------------------------*/
/*--------------------------------------------------------------------*
 *  *---------------------------------------------------------------- *
 *  * Public Functions                                              * *
 *  *---------------------------------------------------------------- *
 *--------------------------------------------------------------------*/
short  DXProcessInit              (void);
void   DXProcessTerm              (void);
short  DXThreadAct                (void);
void   DXThreadTerm               (void);
short  DXExtract(UCHAR, DX_PARMS *, DXCA *);
short  DXGetActColDefs(UCHAR,             /* DBMS Type                */
                      MEM_HEAP_HANDLE,    /* MemHeap to alloc from    */
                      RDB_COL_LIST **,    /* Rtned Process ColList    */
                      RDB_COL_LIST **);   /* Rtned Table ColList      */


/* Internal Prototypes */
short  InitDXTableAndRels(PST_THANDLE, DXCB *, DX_PARMS *, DX_INPUT_TABLE *);
short  InitArchiveActionsAndIndexes(PST_THANDLE, DXCB *, DX_PARMS *);
short  DetermineExtractStrategy(PST_THANDLE, DXCB *);
short  MarkUnTravTblsAndRels(PST_THANDLE, DXCB *, DX_PARMS *);
short  BldExplainStructs(PST_THANDLE, DXCB *, MEM_HEAP_HANDLE,
                         DXPL_MAIN **);
short  FinishInitPhase1(PST_THANDLE, DXCB *);
short  InitFileAttachments(PST_THANDLE, DXCB *);
#ifdef SYNCPOINT
short  PrepareCompiledObject(PST_THANDLE, DX_PARMS *, DXCB *);
short  RestoreCompiledObject(PST_THANDLE, DX_PARMS *, DXCB **,
                             MEM_HEAP_HANDLE);
#endif
short  InitPhase2(PST_THANDLE, DXCB *, DX_PARMS *);
short  ArchiveInitPhase2(PST_THANDLE, DXCB *, DX_PARMS *, long);
short  BldStdSELECTS(PST_THANDLE, DXCB *);
short  BuildPKSelect(PST_THANDLE, DXCB *, DX_TABLE *, DX_PK *);
short  DXBldWhereNullSELECT(PST_THANDLE, DXCB *, DX_TABLE *, void *,
                            short, RDB_COL_LIST *, COL_XR_TOKEN *,
                            UCHAR *, RDB_COL_LIST **, WCHAR **, long *);
short  BldHintedSelect(PST_THANDLE  pThrd,
                       DXCB *       pDXCB,
                       DX_TABLE *   pDXTable);
void   BldTblTok(wchar_t *, DX_TABLE *);
void   BldRelTok(wchar_t *, DX_REL *);
short  DXProcessStartTable(PST_THANDLE, DXCB *);
short  DXProcessCycleCell(PST_THANDLE, DXCB *, DX_CELL *, short);
short  DXProcessTable(PST_THANDLE, DXCB *, DX_TABLE *, short);
short  DXProcessRefTable(PST_THANDLE, DXCB *, DX_TABLE *);
short  DXStepCleanup(PST_THANDLE, DXCB *, short);
short  DXProcessARow(PST_THANDLE, DXCB *, DX_TABLE *, UCHAR *,
                    short, short, short,DX_WORK_AREA*);
void   DXStatusFilePrint(PST_THANDLE, UCHAR *);
short  CheckForAvailIndex(PST_THANDLE, DXCB *, DX_TABLE *,
                          COL_XR_TOKEN *, UCHAR *, UCHAR *);
void  BuildSQLStatsStruct(PST_THANDLE   pThrd,
                          DXCB *        pDXCB,
                          short         AccessType,
                          long          ElapsedTime,
                          long          CompoundKeyCount,
                          wchar_t *     pRelName,
                          DX_WORK_AREA *pStructID);
short  TraceSQLStmt(PST_THANDLE, WCHAR *, long);
short  UpdateActionHVRec(PST_THANDLE     pThrd,
                         DXCB *          pDXCB,
                         DX_TABLE *      pDXTable,
                         long            ActionType,
                         UCHAR           DBMSType);
short  BuildPathColList(PST_THANDLE      pThrd,
                        DXCB *           pDXCB,
                        COL_XR_TOKEN *   p1stToken,
                        DX_PATH_COLS *   pPathCols);

short BuildPKSelectWithINPred(PST_THANDLE     pThrd,
                              DXCB *          pDXCB,
                              DX_TABLE *      pDXTable,
                              DX_PK *         pDXPK,
                              short           INCount,
                              wchar_t **      ppStmt);
short BuildFKSelectWithINPred(PST_THANDLE     pThrd,
                              DXCB *          pDXCB,
                              DX_TABLE *      pDXTable,
                              DX_REL *        pDXRel,
                              short           INCount,
                              wchar_t **      ppStmt);
short ConcatINCond(PST_THANDLE, wchar_t *, DX_TABLE *,
                   short, short, COL_XR_TOKEN *, RDB_COL_LIST *);

/* Funcs to simulate CSS when doing an extract of an extract */
short  DX_CSSGetTblHandle (PST_THANDLE        pThrd,
                           DXCB *             pDXCB,
                           wchar_t *          pDBAliasName,
                           wchar_t *          pCreatorID,
                           wchar_t *          pTableName,
                           RDB_TBL_DEF *      pTblDef,
                           XFTABLE **         ppRtnXFTable);

short DX_CSSTblCols (PST_THANDLE         pThrd,
                     DXCB *              pDXCB,
                     RDB_TBL_DEF *       pTblDef,
                     RDB_COL_LIST **     ppColList,
                     DX_TABLE *          pDXTable);

short DX_CSSColAttr (PST_THANDLE         pThrd,
                     DXCB *              pDXCB,
                     RDB_TBL_DEF *       pTblDef,
                     CSS_COL_ATTR **     ppColAttr,
                     RDB_MEM_HANDLE *    phRDBMem);

short DX_CSSTblPK (PST_THANDLE         pThrd,
                   DXCB *              pDXCB,
                   RDB_TBL_DEF *        pTblDef,
                   CSS_PK *             pPK);


#define PST_DXX_LOCK                            \
{                                               \
    if(pWorkArea &&                             \
       pWorkArea->fParallel && !fLocked){       \
       short rc;                                \
       rc = ENV_ACQUIRE_LOCK(pDXCB->LockTok);   \
       if(rc!=PST_SUCCESS)                      \
          PST_GOTO_UNEXPECTED_RC(pThrd, Lbl, rc,\
                         "ENV_ACQUIRE_LOCK");   \
       fLocked=TRUE;                            \
    }                                           \
}


#define PST_DXX_UNLOCK                          \
{                                               \
    if(pWorkArea &&                             \
       pWorkArea->fParallel &&  fLocked){       \
       short rc;                                \
       rc = ENV_RELEASE_LOCK(pDXCB->LockTok);   \
       if(rc!=PST_SUCCESS)                      \
          PST_GOTO_UNEXPECTED_RC(pThrd, Lbl, rc,\
                         "ENV_RELEASE_LOCK");   \
       fLocked=FALSE;                           \
    }                                           \
}


#define PST_LOCK_GOTO_UNEXPECTED_RC(a,b,c,d)    \
{                                               \
    if(pWorkArea &&                             \
       pWorkArea->fParallel && !fLocked){       \
       ENV_ACQUIRE_LOCK(pDXCB->LockTok);        \
       fLocked=TRUE;                            \
    }                                           \
    PST_GOTO_UNEXPECTED_RC(a, b, c, d);         \
}

#define PST_LOCK_GOTO_RETURN_ERR(a,b,c)         \
{                                               \
    if(pWorkArea &&                             \
       pWorkArea->fParallel && !fLocked){       \
       ENV_ACQUIRE_LOCK(pDXCB->LockTok);        \
       fLocked=TRUE;                            \
    }                                           \
    PST_GOTO_RETURN_ERR(a, b, c);               \
}

#define PST_LOCK_GOTO_RETURN_STATUS(a,b,c)      \
{                                               \
    if(pWorkArea &&                             \
       pWorkArea->fParallel && !fLocked){       \
       ENV_ACQUIRE_LOCK(pDXCB->LockTok);        \
       fLocked=TRUE;                            \
    }                                           \
    PST_GOTO_RETURN_STATUS(a, b, c);            \
}

#define PST_UNLOCK_GOTO_UNEXPECTED_RC(a,b,c,d)  \
{                                               \
    if(pWorkArea &&                             \
       pWorkArea->fParallel && fLocked){        \
       ENV_RELEASE_LOCK(pDXCB->LockTok);        \
       fLocked=FALSE;                           \
    }                                           \
    PST_GOTO_UNEXPECTED_RC(a, b, c, d);         \
}

#define PST_UNLOCK_GOTO_RETURN_ERR(a,b,c)       \
{                                               \
    if(pWorkArea &&                             \
       pWorkArea->fParallel && fLocked){        \
       ENV_RELEASE_LOCK(pDXCB->LockTok);        \
       fLocked=FALSE;                           \
    }                                           \
    PST_GOTO_RETURN_ERR(a, b, c);               \
}

#define PST_UNLOCK_GOTO_RETURN_STATUS(a,b,c)    \
{                                               \
    if(pWorkArea &&                             \
       pWorkArea->fParallel && fLocked){        \
       ENV_RELEASE_LOCK(pDXCB->LockTok);        \
       fLocked=FALSE;                           \
    }                                           \
    PST_GOTO_RETURN_STATUS(a, b, c);            \
}

#define  DX_UNLOCK_CHECK_STOP(pCB)                                     \
{   if (pCB->ErrCode) {                                                \
        PST_UNLOCK_GOTO_RETURN_ERR(pThrd, Lbl, DX_STOP);               \
    }                                                                  \
    if (pCB->pAbortSemaphore != NULL && *pCB->pAbortSemaphore != 0) {  \
        pCB->ErrCode = DX_CANCEL;                                      \
        PST_UNLOCK_GOTO_RETURN_ERR(pThrd, Lbl, DX_STOP);               \
}   }

#define  DX_LOCK_CHECK_STOP(pCB)                                       \
{   if (pCB->ErrCode) {                                                \
        PST_LOCK_GOTO_RETURN_ERR(pThrd, Lbl, DX_STOP);                 \
    }                                                                  \
    if (pCB->pAbortSemaphore != NULL && *pCB->pAbortSemaphore != 0) {  \
        pCB->ErrCode = DX_CANCEL;                                      \
        PST_LOCK_GOTO_RETURN_ERR(pThrd, Lbl, DX_STOP);                 \
}   }

#define DX_PARALLEL_THRESHOLD 5000
#define DX_PARALLEL_FK_FACT   3


#ifdef PST_CPLUS
}
#endif
#endif
