<?xml version="1.0" encoding="ISO-8859-1"?>
<xsl:stylesheet version="1.0"
	xmlns:xsl="http://www.w3.org/1999/XSL/Transform" xmlns:optim="http://www.ibm.com/optim/audit/v11.3.0.0"
	xmlns:optimreport="http://www.ibm.com/optim/report/v11.3.0.0"
	xmlns:optimcommon="http://www.ibm.com/optim/common/v11.3.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
	<xsl:import href="CommonTemplates.xsl"/>
	<xsl:output method="html" indent="yes" version="4.0"
		doctype-public="-//W3C//DTD HTML 4.01//EN" doctype-system="http://www.w3.org/TR/html4/strict.dtd" />

	<!-- BEGIN DBMS access statistics processing -->
	<xsl:template match="optimreport:dbmsAccess">
		<!-- Minor section header -->
		<div>
			<h3 class="table_of_contents_header_text">
				<xsl:text>DBMS Access</xsl:text>
			</h3>
		</div>

		<div class="statistics_container">
			
			<!-- First table in DBMS Section - 3 fields  -->
			<table class="statistics_table">
				<tr>
					<th>
						<xsl:text>Access type</xsl:text>
					</th>
					<td>
						<xsl:call-template name="stringCheck">
							<xsl:with-param name="value" select="@accessType"/>
						</xsl:call-template>
					</td>
				</tr>
				<tr>
					<th>
						<xsl:text>Keys per cursor</xsl:text>
					</th>
					<td>
						<xsl:call-template name="numberFormat">
							<xsl:with-param name="number" select="@keysPer"/>
							<xsl:with-param name="digitsAfterPoint">0</xsl:with-param>
						</xsl:call-template>
					</td>
				</tr>
				<tr>
					<th>
						<xsl:text>Open cursors</xsl:text>
					</th>
					<td>
						<xsl:call-template name="numberFormat">
							<xsl:with-param name="number" select="@openCursor"/>
							<xsl:with-param name="digitsAfterPoint">0</xsl:with-param>
						</xsl:call-template>
					</td>
				</tr>
			</table>

			<!-- Second table in DBMS Section - 4 fields  -->
			<table class="statistics_table">
				<tr class="rightAlign">
					<th>
						<xsl:text>Rows fetched</xsl:text>
					</th>
					<td>
						<xsl:call-template name="numberFormat">
							<xsl:with-param name="number" select="@rowsFetched"/>
							<xsl:with-param name="digitsAfterPoint">0</xsl:with-param>
						</xsl:call-template>
					</td>
				</tr>
				<tr class="rightAlign">
					<th>
						<xsl:text>Rows per second</xsl:text>
					</th>
					<td>
						<xsl:call-template name="numberFormat">
							<xsl:with-param name="number" select="@rowsPerSecond"/>
							<xsl:with-param name="digitsAfterPoint">0</xsl:with-param>
						</xsl:call-template>
					</td>
				</tr>
				<tr class="rightAlign">
					<th>
						<xsl:text>Rows written</xsl:text>
					</th>
					<td>
						<xsl:call-template name="numberFormat">
							<xsl:with-param name="number" select="@rowsWritten"/>
							<xsl:with-param name="digitsAfterPoint">0</xsl:with-param>
						</xsl:call-template>
					</td>
				</tr>
			</table>

			<!-- Third table in DBMS Section - 3 Sections -->
			<table class="statistics_table">
				<tr>
					<th>
						<xsl:text>Process time</xsl:text>
					</th>
					<td>
						<!-- We are getting process time as seconds, we need to format it. -->
						<xsl:variable name="seconds" select="@processTime"/>
						<xsl:variable name="catDate">
							<xsl:value-of select="substring-before($seconds, '.')"/>
						</xsl:variable>

						<xsl:choose>
							<xsl:when test="$catDate">
								<xsl:value-of select="$catDate"/>
							</xsl:when>
							<xsl:otherwise>
								<xsl:value-of select="$seconds"/>
							</xsl:otherwise>
						</xsl:choose>
					</td>
				</tr>
				<tr>
					<th>
						<xsl:text>Time in actions</xsl:text>
					</th>
					<td>
						<xsl:choose>
							<xsl:when test="number(@timeInActions) = number(@timeInActions)">
								<xsl:value-of select="concat(@timeInActions, '%')" />
							</xsl:when>
							<xsl:otherwise>
								<xsl:text>0%</xsl:text>
							</xsl:otherwise>
						</xsl:choose>
					</td>
				</tr>
				<tr>
					<th>
						<xsl:text>Time in database</xsl:text>
					</th>
					<td>
						<xsl:choose>
							<xsl:when test="number(@timeInDatabase) = number(@timeInDatabase)">
								<xsl:value-of select="concat(@timeInDatabase, '%')" />
							</xsl:when>
							<xsl:otherwise>
								<xsl:text>0%</xsl:text>
							</xsl:otherwise>
						</xsl:choose>
					</td>
				</tr>
			</table>
		</div>
	</xsl:template>
	<!-- END DBMS access statistics processing -->

</xsl:stylesheet>

