/*
* -cr_beg-
* ********************************************************************
* IBM Confidential
* OCO Source Materials
* 5724-U38, 5724-U39, 5724-U48, 5655-V17, 5655-V19, 5655-V35, 5655-V37
* (C) Copyright IBM Corp. 2012, 2013
* The source code for this program is not published or otherwise
* divested of its trade secrets, irrespective of what has
* been deposited with the U.S. Copyright Office.
* ********************************************************************
* -cr_end-
*/
/*
 * RTC/Jazz change log
 * 
 * 
 * 15 Jul 2013 Shilpa 9.1.0.5 WI 52348
 * Added new Class for Optim Report Intiative instantiated for each Action
 *
 *

/*!\file
 *\verbatim
 *--------------------------------------------------------------------*
 * File:    OptimDistReportJNI.c
 * Desc:     Interface between Optim Distributed C/C++ reporting and Java
 *  JAXB classes
 *
 * Function: This file contains the following functions:
 *
 *
 *
 * Author:   RAP
 * Created:  6/16/13
 *
 *
 *--------------------------------------------------------------------*
\endverbatim */

#include "RtStdAfx.h"
#include "RtCommon.h"       // Our common includes
#include <jni.h>
#include "OptimReportingContext.h"

/* There is a strange #endif problem in RtStdAfx.h. So comment this out for TEST_MODE */


static const char *classPathOption = "-Djava.class.path=";
static const char *classBasePath = "com/ibm/nex/xml/schema";
static const char *reportingJarName = "optimdistreport.jar";
static const char *OptimReportingContextClassName = "com/ibm/nex/xml/schema/OptimReportingContext";
static const char *OptimReportHelperClassName = "com/ibm/nex/xml/schema/reporting/OptimReportHelper";
static const char *OptimReportingMessageClassName = "com/ibm/nex/xml/schema/OptimReportingMessage";
static const char *OptimReportingNameValuePairClassName = "com/ibm/nex/xml/schema/OptimReportingNameValuePair";
static const char *OptimReportingNumberPairClassName = "com/ibm/nex/xml/schema/OptimReportingNumberPair";
static const char *OptimReportingEntityClassName = "com/ibm/nex/xml/schema/OptimReportingEntity";
static const char *OptimReportingRelationshipClassName = "com/ibm/nex/xml/schema/OptimReportingRelationship";
static const char *OptimReportingStatisticsClassName = "com/ibm/nex/xml/schema/OptimReportingStatistics";

static const int RC_SUCCESS = 0;
static const int RC_WARNING = 4;
static const int RC_ERROR   = 8;
static const int RC_FATAL   = 12;

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#if defined(WIN32)
static const char pathSepChar = '\\';
static const char *pathSepString = "\\";
#else
static const char pathSepChar = '/';
static const char *pathSepString = "/";
#endif

static char *getBasename(char *progStr);
static char *truncateBasename(char *progStr);
static void errorMsg(char *fmtStr, ...);
static void debugMsg(char *fmtStr, ...);

static jstring convertWcharToJString(wchar_t *wsource, JNIEnv *env);
static jstring convertWcharAToJString(wchar_t wSource[ORC_MAX_LEN], JNIEnv *env);

static unsigned char debugFlag = 1;

#if !defined(Convert32To16)
static int Convert32To16(const unsigned int *Utf32,
                     int Utf32Chars,
                     unsigned short *Utf16,
                     int Utf16Size,
                     int *pOutUtf16Size);
#endif

// Functions
int initJvm(JavaVM **javaVM, JNIEnv **env);
int createReport(jobject javaContext, JavaVM *javaVM, JNIEnv *env);
int populateJavaContext(OptimReportingContext_t *context, jobject *javaContext, JavaVM *javaVM, JNIEnv *env);
int outputReport(OptimReportingContext_t *context);



//#if !defined(strrchr)
#if 0
char * strrchr(const char *inStr, int c)
{
	char *retPtr = (char *)NULL;
	char *cPtr = inStr;
	int cVal;

	if (inStr == (char *)NULL) {
		return ((char *)NULL);
	}

	while ((cVal = *cPtr++) != '\0')
	{
		if (cVal == c) {
			retPtr = cPtr - 1; // back the pointer down, since we incremented
		}
	}
	return (retPtr);
}
#endif

/*!\verbatim
 *--------------------------------------------------------------------*
 * Func:     initJvm
 *
 * Purpose:  Initializes JVM and activates reporting jar.
 *
 * Parms:
 *     JavaVM** pointer to java VM pointer -- java VM pointer populated
 *              by call
 *	   JNIEnv** pointer to JNI environment pointer -- JNI environment
 *	            pointer populated by call
 *
 * Returns:
 *		Return code from Java method invoked.
 *
 *--------------------------------------------------------------------*
\endverbatim */

int initJvm(JavaVM **javaVM, JNIEnv **env)
{
    JavaVMOption options[1];
    char optionString[1024];
    //char programName[1024]; // Overkill, but consistent with optionString
    char *baseProgramDir;
    JavaVMInitArgs initArgs;
    int rc;

    (void)strcpy(optionString, classPathOption);
//    (void) strcat(optionString, argv[0]);
//    (void)strcpy(programName,argv[0]);
//    baseProgramDir = truncateBasename(programName);
    //baseProgramDir = ".";
     baseProgramDir = "D:\\RTWIN\\910\\Source\\DebugUni";
    (void)strcat(optionString,baseProgramDir);
    (void)strcat(optionString,pathSepString);
    (void)strcat(optionString,reportingJarName);
    options[0].optionString = optionString;

    initArgs.version = JNI_VERSION_1_6;
    initArgs.nOptions = 1;
    initArgs.options = options;
    initArgs.ignoreUnrecognized = JNI_TRUE;
    debugMsg("JNI options are:  %s\n", options[0].optionString);
    rc = JNI_CreateJavaVM(javaVM, (void**)env, &initArgs);
    if (rc < 0) {
        errorMsg("Error: Unable to create JVM.\n");
        return RC_FATAL;
    }

    return rc;
}

int createReport(jobject javaContext, JavaVM *javaVM, JNIEnv *env)
{
	int rc = RC_SUCCESS;
	jclass jOptimReportHelperClass;
	jmethodID jReportCreate;

	/* Call OptimReportHelper.OptimReportHelper.createProcessAndDisplayReportFromContext with the populated context */
	jOptimReportHelperClass = env->FindClass(OptimReportHelperClassName);
    if(env->ExceptionCheck()) {
        env->ExceptionDescribe();
        env->ExceptionClear();
        /* Fall through, since an exception will result in a NULL return */
    }
    if (jOptimReportHelperClass == NULL) {
        errorMsg("Error: Unable to find report helper class. %s\n", OptimReportHelperClassName);
        return RC_FATAL;
    }
    jReportCreate = env->GetStaticMethodID(jOptimReportHelperClass,"createProcessAndDisplayReportFromContext", "(Lcom/ibm/nex/xml/schema/OptimReportingContext;)I");
    if(env->ExceptionCheck())  {
        env->ExceptionDescribe();
        env->ExceptionClear();
        /* Fall through, since an exception will result in a NULL return */
    }
    if (jReportCreate == NULL) {
        errorMsg("Error: Unable to find OptimReportHelper.createProcessAndDisplayReportFromContext() method.");
        return RC_FATAL;
    }

    rc = env->CallStaticIntMethod(jOptimReportHelperClass, jReportCreate, javaContext);
    if(env->ExceptionCheck()) {
        env->ExceptionDescribe();
        env->ExceptionClear();
        errorMsg("Error: Exception while invoking OptimReportHelper.createProcessAndDisplayReportFromContext() method.");
        rc = RC_FATAL;
    }
	return rc;
}

#define POPULATE_STRING_FIELD(fieldName, setterName, javaClass, javaObject, container) \
	    if (container fieldName != (wchar_t *)NULL) { \
	    	javaStringValue = convertWcharToJString(container fieldName, env); \
	    	if (javaStringValue != (jstring) NULL) { \
	        	javaSetter = env->GetMethodID(javaClass, #setterName, "(Ljava/lang/String;)V"); \
	        	if (javaSetter != (jmethodID) NULL) { \
	        		env->CallVoidMethod(javaObject, javaSetter, javaStringValue); \
	        	    if(env->ExceptionCheck()) { \
	        	        env->ExceptionDescribe(); \
	        	        env->ExceptionClear(); \
	        	        errorMsg("Error: Exception while invoking %s method.", #setterName); \
	        	        return RC_FATAL; \
	        	    } \
	                /*env->DeleteLocalRef(javaSetter);*/ \
	        	} else { \
	        		errorMsg("Could not find method %s\n", #setterName); \
	        		env->DeleteGlobalRef(javaStringValue); \
	        		return RC_FATAL; \
	        	} \
	        	env->DeleteGlobalRef(javaStringValue); \
	    	} else { \
	    		errorMsg("Could not obtain string value for field %s\n", #fieldName); \
	    		return RC_FATAL; \
	    	} \
	    }

#define POPULATE_NUMBER_FIELD(fieldName, setterName, typeSig, castType, javaClass, javaObject, container) \
	    { \
		    javaSetter = env->GetMethodID(javaClass, #setterName, typeSig); \
		    if (javaSetter != (jmethodID) NULL) { \
				env->CallVoidMethod(javaObject, javaSetter, (castType)container fieldName); \
        	    if(env->ExceptionCheck()) { \
        	        env->ExceptionDescribe(); \
        	        env->ExceptionClear(); \
        	        errorMsg("Error: Exception while invoking %s method.", #setterName); \
        	        return RC_FATAL; \
        	    } \
		        /*env->DeleteLocalRef(javaSetter);*/ \
		    } else { \
				errorMsg("Could not find method %s\n", #fieldName); \
				return RC_FATAL; \
		    } \
		}

#define POPULATE_INT_FIELD(fieldName, setterName, javaClass, javaObject, container) POPULATE_NUMBER_FIELD(fieldName, setterName, "(I)V", jint, javaClass, javaObject, container)
#define POPULATE_LONG_FIELD(fieldName, setterName, javaClass, javaObject, container) POPULATE_NUMBER_FIELD(fieldName, setterName, "(J)V", jlong, javaClass, javaObject, container)
#define POPULATE_BOOLEAN_FIELD(fieldName, setterName, javaClass, javaObject, container) POPULATE_NUMBER_FIELD(fieldName, setterName, "(Z)V", jboolean, javaClass, javaObject, container)
#define POPULATE_DOUBLE_FIELD(fieldName, setterName, javaClass, javaObject, container) POPULATE_NUMBER_FIELD(fieldName, setterName, "(D)V", jdouble, javaClass, javaObject, container)

#define POPULATE_CONTEXT_INT_FIELD(fieldName, setterName) POPULATE_NUMBER_FIELD(fieldName, setterName, "(I)V", jint, jContextCls, localContext, context->)
#define POPULATE_CONTEXT_LONG_FIELD(fieldName, setterName) POPULATE_NUMBER_FIELD(fieldName, setterName, "(J)V", jlong, jContextCls, localContext, context->)
#define POPULATE_CONTEXT_BOOLEAN_FIELD(fieldName, setterName) POPULATE_NUMBER_FIELD(fieldName, setterName, "(Z)V", jboolean, jContextCls, localContext, context->)
#define POPULATE_CONTEXT_DOUBLE_FIELD(fieldName, setterName) POPULATE_NUMBER_FIELD(fieldName, setterName, "(D)V", jdouble, jContextCls, localContext, context->)
#define POPULATE_CONTEXT_STRING_FIELD(fieldName, setterName) POPULATE_STRING_FIELD(fieldName, setterName, jContextCls, localContext, context->)

#define CONFIGURE_LIST_ACCESS(listAccessMethodName,fullClassName,constructorSignature) \
		{\
    jListGetter = env->GetMethodID(jContextCls, #listAccessMethodName, "()Ljava/util/List;"); \
    jList = (jobject) NULL; /* Clear out jList */	\
	if (jListGetter != (jmethodID) NULL) {	\
		jList = env->CallObjectMethod(localContext, jListGetter); \
        /*env->DeleteLocalRef(javaSetter);*/ \
	} else { \
		errorMsg("Could not find method %s\n", #listAccessMethodName); \
		return RC_FATAL; \
	} \
	/* Set things up to create element objects */ \
	debugMsg("LIST ops set up, now set up %s object stuff\n", fullClassName); \
	jListElementClass = env->FindClass(fullClassName); \
    if (jListElementClass == NULL) { \
        errorMsg("Error: Unable to find list element class. %s\n", fullClassName); \
        return RC_FATAL; \
    }\
    jListElementConstructor = env->GetMethodID(jListElementClass, "<init>", constructorSignature);\
    if (jListElementConstructor == NULL) { \
        errorMsg("Error: Unable to find %s constructor.", fullClassName); \
        return RC_FATAL; \
    } \
		}

#define ADD_LIST_ELEMENT(fullClassName) \
{ \
    addStatus = env->CallBooleanMethod(jList,jArrayListAddMID, jListElement); \
    if(env->ExceptionCheck()) { \
        env->ExceptionDescribe(); \
        env->ExceptionClear(); \
    	errorMsg("Error: add() of element to list %s generated Exception.\n", fullClassName);\
    	return RC_FATAL; \
    } \
    if (!addStatus) { \
    	errorMsg("Error: add() of element to list %s failed.\n", fullClassName); \
    	return RC_FATAL; \
    } \
}


int populateJavaContext(OptimReportingContext_t *context, jobject *javaContext, JavaVM *javaVM, JNIEnv *env)
{
	int rc = RC_SUCCESS;
	jclass jContextCls;
	jmethodID jContextConstructor;
	jobject localContext;
	jmethodID javaSetter;
	jstring  javaStringValue;
	jstring	 javaStringValue2;
	jclass stringClass;
	jclass jArrayListClass;
    jmethodID jArrayListAddMID;
    jmethodID jListGetter;
    jobject   jList;
    jboolean  addStatus;
    jclass	  jListElementClass;
    jmethodID jListElementConstructor;
    jobject   jListElement;
	int idx;

    debugMsg("Finding String class in populate\n");
    stringClass = env->FindClass("java/lang/String");
    if (stringClass == NULL) {
          errorMsg("Error: Unable to find String\n");
    } else {
    	debugMsg("Found String!!\n");
    }

	jContextCls = env->FindClass(OptimReportingContextClassName);
    if (jContextCls == NULL) {
        errorMsg("Error: Unable to find report context class. %s\n", OptimReportingContextClassName);
        return RC_FATAL;
    }
    jContextConstructor = env->GetMethodID(jContextCls, "<init>", "()V");
    if (jContextConstructor == NULL) {
        errorMsg("Error: Unable to find OptimReportingContext constructor.");
        return RC_FATAL;
    }

    localContext = env->NewObject(jContextCls, jContextConstructor);
    if (localContext == NULL) {
        errorMsg("Error: Unable to instantiate loader.\n");
        return RC_FATAL;
    }
    *javaContext = env->NewGlobalRef(localContext);

    POPULATE_CONTEXT_STRING_FIELD(reportFile, setReportFileName)             /* reportFile */
    POPULATE_CONTEXT_STRING_FIELD(htmlReportFile, setHtmlReportFileName)     /* htmlReportFile */

    /* It would be nice to do some fancy reflection thing here.  BUT, for now, brute force will do */
    /* AccessDefinitionModificationDate */
    POPULATE_CONTEXT_LONG_FIELD(errorCount, setErrorCount)
    POPULATE_CONTEXT_LONG_FIELD(infoCount, setInfoCount)
    POPULATE_CONTEXT_LONG_FIELD(warningCount, setWarningCount)
    POPULATE_CONTEXT_LONG_FIELD(timeElapsed, setTimeElapsed)
    POPULATE_CONTEXT_LONG_FIELD(totalRows, setTotalRows)
    POPULATE_CONTEXT_LONG_FIELD(totalRowsFirstPass, setTotalRowsFirstPass)
    POPULATE_CONTEXT_LONG_FIELD(totalRowsInError, setTotalRowsInError)
    POPULATE_CONTEXT_LONG_FIELD(totalTables, setTotalTables)

    POPULATE_CONTEXT_STRING_FIELD(optimDirectoryName, setOptimDirectoryName)
    POPULATE_CONTEXT_STRING_FIELD(optimDirectoryDBMSType, setOptimDirectoryDBMSType)
    POPULATE_CONTEXT_STRING_FIELD(optimDirectoryDBMSVendor, setOptimDirectoryDBMSVendor)
    POPULATE_CONTEXT_STRING_FIELD(optimDirectoryDBMSVersion, setOptimDirectoryDBMSVersion)
    POPULATE_CONTEXT_STRING_FIELD(optimSolutionName, setOptimSolutionName)
    POPULATE_CONTEXT_STRING_FIELD(optimSolutionVersion, setOptimSolutionVersion)
    POPULATE_CONTEXT_STRING_FIELD(optimRuntimeName, setOptimRuntimeName)
    POPULATE_CONTEXT_STRING_FIELD(optimRuntimeVersion, setOptimRuntimeVersion)
    POPULATE_CONTEXT_STRING_FIELD(clientUserId, setClientUserId)
    POPULATE_CONTEXT_STRING_FIELD(requestName, setRequestName)
    POPULATE_CONTEXT_STRING_FIELD(serverName, setServerName)
    POPULATE_CONTEXT_STRING_FIELD(serverUserId, setServerUserId)
    POPULATE_CONTEXT_STRING_FIELD(timeStarted, setTimeStarted)
    POPULATE_CONTEXT_STRING_FIELD(timeFinished, setTimeFinished)
    POPULATE_CONTEXT_STRING_FIELD(accessDefinitionName, setAccessDefinitionName)
    POPULATE_CONTEXT_STRING_FIELD(accessDefinitionModificationDate, setAccessDefinitionModificationDate)
    POPULATE_CONTEXT_STRING_FIELD(extractFile, setExtractFile)
    POPULATE_CONTEXT_STRING_FIELD(compareFile, setCompareFile)
    POPULATE_CONTEXT_STRING_FIELD(controlFile, setControlFile)    

    POPULATE_CONTEXT_INT_FIELD(accessDefinitionType, setAccessDefinitionType)
    POPULATE_CONTEXT_INT_FIELD(reportType, setReportType) 
    POPULATE_CONTEXT_STRING_FIELD(tableMapName, setTableMapName)   
    /*
    POPULATE_CONTEXT_STRING_FIELD(ddl, setDDL) 
    POPULATE_CONTEXT_STRING_FIELD(dataSource1, setDataSource1)                             
    POPULATE_CONTEXT_STRING_FIELD(dataSource2, setDataSource2)                             
    POPULATE_CONTEXT_INT_FIELD(nonEqualSets, setNonEqualSets)                                
    POPULATE_CONTEXT_INT_FIELD(singleTable, setSingleTable)                                
    POPULATE_CONTEXT_INT_FIELD(sourceType, setSourceType)                                  
    POPULATE_CONTEXT_INT_FIELD(sourceType2, setSourceType2)                                 
    POPULATE_CONTEXT_INT_FIELD(totalSets, setTotalSets)                                   
    POPULATE_CONTEXT_INT_FIELD(numberOfArchiveFiles, setNumberofArchiveFiles)                        
    POPULATE_CONTEXT_INT_FIELD(numberOfSubsetFiles, setNumberofSubsetFiles)                         
    POPULATE_CONTEXT_INT_FIELD(numberOfFilesRestored, setNumberofFilesRestored)                       
    POPULATE_CONTEXT_STRING_FIELD(reportName, setReportName)                             
    POPULATE_CONTEXT_STRING_FIELD(reportRole, setReportRole)                             
    POPULATE_CONTEXT_STRING_FIELD(reportState, setReportState)                             
    
    POPULATE_CONTEXT_INT_FIELD(importImported, setImportImported)                             
    POPULATE_CONTEXT_STRING_FIELD(fileName, setFileName)                                
    POPULATE_CONTEXT_LONG_FIELD(fileSize, setFileSize) 
    */
    
    /* Now populate the lists... First -- configure JNI for handling ArrayList */
    debugMsg("Going after Messages list\n");
    jArrayListClass = env->FindClass("java/util/ArrayList");
    jArrayListAddMID = env->GetMethodID(jArrayListClass, "add", "(Ljava/lang/Object;)Z");
    /* Get the messages list */
    CONFIGURE_LIST_ACCESS(getMessages,OptimReportingMessageClassName,"(JLjava/lang/String;I)V")
	debugMsg("Got message object stuff done. Populating\n");
    for (idx = 0; idx < context->numMessages; idx ++) {      
      /* Do String construction/conversion */
    	javaStringValue = convertWcharAToJString(context->pMessages->message, env);
    	if (javaStringValue == (jstring) NULL) {
    		errorMsg("Could not obtain string value for field %s\n", "reportFile");
    		return RC_FATAL;
    	}
    	/* Create list element */
        jListElement = env->NewObject(jListElementClass, jListElementConstructor,(jlong)context->pMessages->id, javaStringValue , (jint)context->pMessages->type);
        if (jListElement == NULL) {
            errorMsg("Error: Unable to instantiate element for %s.\n", OptimReportingMessageClassName);
            return RC_FATAL;
        }
        /* Add element to list and clean up references */
        ADD_LIST_ELEMENT(OptimReportingMessageClassName)
        env->DeleteLocalRef(jListElement);
        env->DeleteGlobalRef(javaStringValue);
        context->pMessages = context->pMessages->pNext;
    }
   	env->DeleteGlobalRef(jList);
   	env->DeleteGlobalRef(jListElementClass);
    jList = (jobject) NULL; /* Clear out jList */
    jListElementClass = (jclass) NULL;

    debugMsg ("Done with message list \n");
    /* Do the Request Options Next */
    debugMsg ("Beginning Request Options\n");
    /* Get the request options list */
    CONFIGURE_LIST_ACCESS(getRequestOptions,OptimReportingNameValuePairClassName,"(Ljava/lang/String;Ljava/lang/String;)V")
	debugMsg("Got request option object stuff done. Populating\n");
    for (idx = 0; idx < context->numRequestOptions; idx ++) {
    	/* Do String construction/conversion */
    	javaStringValue = convertWcharAToJString(context->pRequestOptions->name, env);
    	if (javaStringValue == (jstring) NULL) {
    		errorMsg("Could not obtain string value for field %s\n", "name");
    		return RC_FATAL;
    	}
    	javaStringValue2 = convertWcharToJString(context->pRequestOptions->value, env);
    	if (javaStringValue2 == (jstring) NULL) {
    		errorMsg("Could not obtain string value for field %s\n", "name");
    		return RC_FATAL;
    	}
    	/* Create list element */
        jListElement = env->NewObject(jListElementClass, jListElementConstructor, javaStringValue, javaStringValue2);
        if (jListElement == NULL) {
            errorMsg("Error: Unable to instantiate element for %s.\n", OptimReportingNameValuePairClassName);
            return RC_FATAL;
        }
        /* Add element to list and clean up references */
        ADD_LIST_ELEMENT(OptimReportingNameValuePairClassName)
        env->DeleteLocalRef(jListElement);
        env->DeleteGlobalRef(javaStringValue);
        env->DeleteGlobalRef(javaStringValue2);
        context->pRequestOptions = context->pRequestOptions->pNext;
    }
   	env->DeleteGlobalRef(jList);
   	env->DeleteGlobalRef(jListElementClass);
    jList = (jobject) NULL; /* Clear out jList */
    jListElementClass = (jclass) NULL;
    debugMsg("Done with RequestOptions\n");

    /* Now do Extended Objects */
    debugMsg ("Beginning ExtendedObjects\n");
    CONFIGURE_LIST_ACCESS(getExtendedObjects,OptimReportingNumberPairClassName,"(JJ)V")
	debugMsg("Got request option object stuff done. Populating\n");
    for (idx = 0; idx < context->numExtendedObjects; idx ++) {
     	/* Create list element */
        jListElement = env->NewObject(jListElementClass, jListElementConstructor, (jlong)context->pExtendedObjects->number1, (jlong)context->pExtendedObjects->number2);
        if (jListElement == NULL) {
            errorMsg("Error: Unable to instantiate element for %s.\n", OptimReportingNumberPairClassName);
            return RC_FATAL;
        }
        /* Add element to list and clean up references */
        ADD_LIST_ELEMENT(OptimReportingNumberPairClassName)
        env->DeleteLocalRef(jListElement);
        context->pExtendedObjects = context->pExtendedObjects->pNext;
    }
   	env->DeleteGlobalRef(jList);
   	env->DeleteGlobalRef(jListElementClass);
    jList = (jobject) NULL; /* Clear out jList */
    jListElementClass = (jclass) NULL;
    debugMsg("Done with ExtendedObjects\n");

    /* CREATE ENTITIES LIST */
    debugMsg ("Beginning Entities\n");
    CONFIGURE_LIST_ACCESS(getEntities,OptimReportingEntityClassName,"()V")
	debugMsg("Got entity object stuff done. Populating\n");
    for (idx = 0; idx < context->numEntities; idx ++) {
     	/* Create list element */
        jListElement = env->NewObject(jListElementClass, jListElementConstructor);
        if (jListElement == NULL) {
            errorMsg("Error: Unable to instantiate element for %s.\n", OptimReportingEntityClassName);
            return RC_FATAL;
        }
        /* Add element to list and clean up references */
        ADD_LIST_ELEMENT(OptimReportingEntityClassName)
        /* NOW, POPULATE FIELDS IN LIST ELEMENT */
        POPULATE_LONG_FIELD(compareDifferent, setCompareDifferent, jListElementClass, jListElement, context->pEntities-> )
        POPULATE_LONG_FIELD(compareDuplicate, setCompareDuplicate, jListElementClass, jListElement, context->pEntities-> )
        POPULATE_LONG_FIELD(compareEqual, setCompareEqual, jListElementClass, jListElement, context->pEntities-> )
        POPULATE_LONG_FIELD(compareUnmatched, setCompareUnmatched, jListElementClass, jListElement, context->pEntities-> )
        POPULATE_STRING_FIELD(convertCsvFileName, setConvertCsvFileName, jListElementClass, jListElement, context->pEntities-> )
        POPULATE_LONG_FIELD(convertCsvLimitFileCount, setConvertCsvLimitFileCount, jListElementClass, jListElement, context->pEntities-> )
        POPULATE_LONG_FIELD(convertCsvLimitFileSize, setConvertCsvLimitFileSize, jListElementClass, jListElement, context->pEntities-> )
        POPULATE_LONG_FIELD(convertCsvLimitRowCount, setConvertCsvLimitRowCount, jListElementClass, jListElement, context->pEntities-> )
        POPULATE_BOOLEAN_FIELD(deleteNotFound, setDeleteNotFound, jListElementClass, jListElement, context->pEntities-> )
        POPULATE_DOUBLE_FIELD(extractAverageRowLength, setExtractAverageRowLength, jListElementClass, jListElement, context->pEntities-> )
        POPULATE_BOOLEAN_FIELD(extractCompressionEnabled, setExtractCompressionEnabled, jListElementClass, jListElement, context->pEntities-> )
        POPULATE_INT_FIELD(extractCompressionValue, setExtractCompressionValue, jListElementClass, jListElement, context->pEntities-> )
        POPULATE_BOOLEAN_FIELD(extractLob, setExtractLob, jListElementClass, jListElement, context->pEntities-> )
        POPULATE_DOUBLE_FIELD(extractMaxLobLength, setExtractMaxLobLength, jListElementClass, jListElement, context->pEntities-> )
        POPULATE_LONG_FIELD(failed, setFailed, jListElementClass, jListElement, context->pEntities-> )
        POPULATE_LONG_FIELD(insertInserted, setInsertInserted, jListElementClass, jListElement, context->pEntities-> )
        POPULATE_LONG_FIELD(insertUpdated, setInsertUpdated, jListElementClass, jListElement, context->pEntities-> )
        POPULATE_STRING_FIELD(name, setName, jListElementClass, jListElement, context->pEntities-> )
        POPULATE_LONG_FIELD(recordCount, setRecordCount, jListElementClass, jListElement, context->pEntities-> )
        POPULATE_LONG_FIELD(skipped, setSkipped, jListElementClass, jListElement, context->pEntities-> )
        POPULATE_STRING_FIELD(status, setStatus, jListElementClass, jListElement, context->pEntities-> )
        POPULATE_INT_FIELD(type, setType, jListElementClass, jListElement, context->pEntities-> )
        env->DeleteLocalRef(jListElement);
        context->pEntities = context->pEntities->pNext;
    }
   	env->DeleteGlobalRef(jList);
   	env->DeleteGlobalRef(jListElementClass);
    jList = (jobject) NULL; /* Clear out jList */
    jListElementClass = (jclass) NULL;
    debugMsg("Done with Entities\n");

    /* Relationships */
    debugMsg ("Beginning Relationships\n");
    CONFIGURE_LIST_ACCESS(getRelationships,OptimReportingRelationshipClassName,"()V")
	debugMsg("Got relationship object stuff done. Populating\n");
    for (idx = 0; idx < context->numRelationships; idx ++) {
     	/* Create list element */
        jListElement = env->NewObject(jListElementClass, jListElementConstructor);
        if (jListElement == NULL) {
            errorMsg("Error: Unable to instantiate element for %s.\n", OptimReportingRelationshipClassName);
            return RC_FATAL;
        }
        /* Add element to list and clean up references */
        ADD_LIST_ELEMENT(OptimReportingRelationshipClassName)
        /* NOW, POPULATE FIELDS IN LIST ELEMENT */
        POPULATE_INT_FIELD(childAccessType, setChildAccessType, jListElementClass, jListElement, context->pRelationships-> )
        POPULATE_INT_FIELD(childKeyLimit, setChildKeyLimit, jListElementClass, jListElement, context->pRelationships-> )
        POPULATE_STRING_FIELD(childName, setChildName, jListElementClass, jListElement, context->pRelationships-> )
        POPULATE_INT_FIELD(parentAccessType, setParentAccessType, jListElementClass, jListElement, context->pRelationships-> )
        POPULATE_INT_FIELD(parentKeyLimit, setParentKeyLimit, jListElementClass, jListElement, context->pRelationships-> )
        POPULATE_STRING_FIELD(parentName, setParentName, jListElementClass, jListElement, context->pRelationships-> )
        POPULATE_STRING_FIELD(relationshipName, setRelationshipName, jListElementClass, jListElement, context->pRelationships-> )
        env->DeleteLocalRef(jListElement);
        context->pRelationships = context->pRelationships->pNext;
    }
   	env->DeleteGlobalRef(jList);
   	env->DeleteGlobalRef(jListElementClass);
    jList = (jobject) NULL; /* Clear out jList */
    jListElementClass = (jclass) NULL;
    debugMsg("Done with Relationships\n");

    /* NOW, do STATISTICS */
    debugMsg ("Beginning Statistics\n");
    CONFIGURE_LIST_ACCESS(getStatistics,OptimReportingStatisticsClassName,"()V")
	debugMsg("Got pStatistics object stuff done. Populating\n");
    for (idx = 0; idx < context->numStatistics; idx ++) {
     	/* Create list element */
        jListElement = env->NewObject(jListElementClass, jListElementConstructor);
        if (jListElement == NULL) {
            errorMsg("Error: Unable to instantiate element for %s.\n", OptimReportingStatisticsClassName);
            return RC_FATAL;
        }
        /* Add element to list and clean up references */
        ADD_LIST_ELEMENT(OptimReportingStatisticsClassName)
        /* NOW, POPULATE FIELDS IN LIST ELEMENT */
        POPULATE_INT_FIELD(childStrategy, setChildStrategy, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_INT_FIELD(columnCount, setColumnCount, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_BOOLEAN_FIELD(cycle, setCycle, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_INT_FIELD(databaseConnections, setDatabaseConnections, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_INT_FIELD(dbAccessType, setDbAccessType, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_LONG_FIELD(dbDataPerSecond, setDbDataPerSecond, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_INT_FIELD(dbKeysPer, setDbKeysPer, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_STRING_FIELD(dbmsVendor, setDbmsVendor, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_STRING_FIELD(dbmsVersion, setDbmsVersion, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_INT_FIELD(dbOpenCursor, setDbOpenCursor, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_LONG_FIELD(dbProcessTime, setDbProcessTime, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_LONG_FIELD(dbRowsFetched, setDbRowsFetched, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_LONG_FIELD(dbRowsPerSecond, setDbRowsPerSecond, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_LONG_FIELD(dbRowsWritten, setDbRowsWritten, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_LONG_FIELD(dbTimeInAct, setDbTimeInAct, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_LONG_FIELD(dbTimeInDB, setDbTimeInDB, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_LONG_FIELD(estimatedRowCount, setEstimatedRowCount, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_LONG_FIELD(fetchBuffer, setFetchBuffer, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_LONG_FIELD(foreignKeyIndex, setForeignKeyIndex, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_LONG_FIELD(foreignKeyNoIndex, setForeignKeyNoIndex, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_BOOLEAN_FIELD(lobs, setLobs, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_INT_FIELD(method, setMethod, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_STRING_FIELD(name, setName, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_INT_FIELD(parentStrategy, setParentStrategy, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_LONG_FIELD(primaryKeyIndex, setPrimaryKeyIndex, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_LONG_FIELD(primaryKeyNoIndex, setPrimaryKeyNoIndex, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_INT_FIELD(relAccess, setRelAccess, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_INT_FIELD(relDb2Cost, setRelDb2Cost, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_INT_FIELD(relDb2ScanCost, setRelDb2ScanCost, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_INT_FIELD(relDirection, setRelDirection, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_BOOLEAN_FIELD(relIndex, setRelIndex, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_INT_FIELD(relKeyLength, setRelKeyLength, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_INT_FIELD(relKeyPerCur, setRelKeyPerCur, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_INT_FIELD(relLookupKey, setRelLookupKey, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_STRING_FIELD(relName, setRelName, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_STRING_FIELD(relSql, setRelSql, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_INT_FIELD(relType, setRelType, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_INT_FIELD(rowLength, setRowLength, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_BOOLEAN_FIELD(selectUniqueRecord, setSelectUniqueRecord, jListElementClass, jListElement, context->pStatistics-> )
        POPULATE_INT_FIELD(stepNumber, setStepNumber, jListElementClass, jListElement, context->pStatistics-> )
        env->DeleteLocalRef(jListElement);
        context->pStatistics = context->pStatistics->pNext;
    }
   	env->DeleteGlobalRef(jList);
   	env->DeleteGlobalRef(jListElementClass);
    jList = (jobject) NULL; /* Clear out jList */
    jListElementClass = (jclass) NULL;
    debugMsg("Done with Statistics\n");



    env->DeleteLocalRef(jContextCls);
    env->DeleteLocalRef(stringClass);
    env->DeleteLocalRef(jArrayListClass);
	return rc;
}

int outputReport(OptimReportingContext_t *context) {
    JavaVM *javaVM;
    JNIEnv *env;
    int rc;
    jobject javaContext;

    /* Check context */
    if (context == (OptimReportingContext_t *) NULL) {
    	errorMsg("Context cannot be NULL.");
    	return RC_FATAL;
    }
    debugMsg("Initializing JVM.\n");

    rc = initJvm(&javaVM, &env);
    if (rc > RC_WARNING) {
    	errorMsg("Problem creating JVM.  Consult logs.\n");
    	return rc;
    }
    /* JVM Created. Need to terminate JVM whenever we bail*/
    /* Now, we need to create an OptimReportingContext in Java and populate it */
    debugMsg("populating Java context \n");
    rc = populateJavaContext(context, &javaContext, javaVM, env);
    if (rc > RC_WARNING) {
    	errorMsg("Problem populating Java OptimReportingContext. Consult logs.\n");
        javaVM->DestroyJavaVM();
        return rc;
    }
    debugMsg("Context populated, outputting report\n");
    /* With the context populated, we pass it to OptimReportHelper.createProcessAndDisplayReportFromContext() */
    rc = createReport(javaContext, javaVM, env);
    if (rc > RC_WARNING) {
    	errorMsg("Problem creating report. Consult logs.\n");
    	/* Fall through to destroy JVM */
    }

    debugMsg("Reporting done RC= %d\n", rc);
    /* Terminate JVM */
    javaVM->DestroyJavaVM();
    return rc;
}

#ifndef WCSLEN
#define WCSLEN(wstr) wcslen((wstr))
#endif
#ifndef WCSNCPY
#define WCSNCPY(wto,wfrom,wlen) wcsncpy((wto),(wfrom),(wlen))
#endif

/* The caller must free the string */
jstring convertWcharToJString(wchar_t *wSource, JNIEnv *env) {
	jstring retStr;
	jstring jDest;
	jobject jGlobRef;
	jchar tmpBuf[256];
	size_t wSourceNumChars;
	jchar *javaTmpStg = (jchar *)NULL;
	jint jStgUsed;

	if (wSource == (wchar_t *) NULL) {
		return ((jstring) NULL);
	}

	wSourceNumChars = WCSLEN(wSource);
	if ((wSourceNumChars+1)*sizeof(jchar) <= sizeof(tmpBuf)) {
		javaTmpStg = tmpBuf;
	} else {
		javaTmpStg = (jchar *)malloc((wSourceNumChars + 1) * sizeof(jchar));
	}
	memset(javaTmpStg, 0, (wSourceNumChars + 1) * sizeof(jchar));

	/* WIDE 4-byte wchar_t */
    if (sizeof(wchar_t) == 4) {
    	Convert32To16((unsigned int *)wSource, wSourceNumChars, javaTmpStg, wSourceNumChars, &jStgUsed);
    } else { /* 2-byte wchar_t -- we accommodate no other choices */
        WCSNCPY((wchar_t *)javaTmpStg, wSource, wSourceNumChars);
        jStgUsed = wSourceNumChars;
    }

 	jDest = env->NewString(javaTmpStg, jStgUsed);
  	if (javaTmpStg != NULL && javaTmpStg != tmpBuf) {
   		free(javaTmpStg);
   	}
   	javaTmpStg = NULL;
    jGlobRef = env->NewGlobalRef(jDest);
    env->DeleteLocalRef(jDest);

    return((jstring)jGlobRef);
}

jstring convertWcharAToJString(wchar_t wSource[ORC_MAX_LEN], JNIEnv *env) {
	jstring retStr;
	jstring jDest;
	jobject jGlobRef;
	jchar tmpBuf[256];
	size_t wSourceNumChars;
	jchar *javaTmpStg = (jchar *)NULL;
	jint jStgUsed;

	if (wSource == (wchar_t *) NULL) {
		return ((jstring) NULL);
	}

	wSourceNumChars = WCSLEN(wSource);
	if ((wSourceNumChars+1)*sizeof(jchar) <= sizeof(tmpBuf)) {
		javaTmpStg = tmpBuf;
	} else {
		javaTmpStg = (jchar *)malloc((wSourceNumChars + 1) * sizeof(jchar));
	}
	memset(javaTmpStg, 0, (wSourceNumChars + 1) * sizeof(jchar));

	/* WIDE 4-byte wchar_t */
    if (sizeof(wchar_t) == 4) {
    	Convert32To16((unsigned int *)wSource, wSourceNumChars, javaTmpStg, wSourceNumChars, &jStgUsed);
    } else { /* 2-byte wchar_t -- we accommodate no other choices */
        WCSNCPY((wchar_t *)javaTmpStg, wSource, wSourceNumChars);
        jStgUsed = wSourceNumChars;
    }

 	jDest = env->NewString(javaTmpStg, jStgUsed);
  	if (javaTmpStg != NULL && javaTmpStg != tmpBuf) {
   		free(javaTmpStg);
   	}
   	javaTmpStg = NULL;
    jGlobRef = env->NewGlobalRef(jDest);
    env->DeleteLocalRef(jDest);

    return((jstring)jGlobRef);
}

char *getBasename(char *progStr)
{
	char *ptr = strrchr(progStr,(int)pathSepChar);
	if (ptr == (char *)NULL) {
		return (progStr);
	}
	return (++ptr);
}

char *truncateBasename(char *progStr)
{
	char *ptr = strrchr(progStr,(int)pathSepChar);
	if (ptr == (char *)NULL) {
		return (NULL);
	}
	*ptr = '\0'; // Truncate
	return (progStr);
}

void errorMsg(char *fmtStr, ...)
{
	va_list argPtr;
	va_start(argPtr, fmtStr);
	/* We can patch into an error mechanism here.  For now, go with fprintf */
	(void)vfprintf(stderr, fmtStr, argPtr);
	va_end(argPtr);
}

void debugMsg(char *fmtStr, ...)
{
	va_list argPtr;
	va_start(argPtr, fmtStr);
	if (debugFlag) {
		/* We can patch into an error mechanism here.  For now, go with fprintf */
		(void)vfprintf(stdout, fmtStr, argPtr);
	}
	va_end(argPtr);
}

#if !(defined Convert32To16)
/*-------------------------------------------------------------------------------------------------
 * Surrogate Defines
 *-------------------------------------------------------------------------------------------------*/
#define SURROGATE_HIGH_START  0xD800   // Surrogate hi word start
#define SURROGATE_HIGH_END    0xDBFF   // "         "  "    end
#define SURROGATE_LOW_START   0xDC00   // "         low word start
#define SURROGATE_LOW_END     0xDFFF   // "         "   "    end

#define SURROGATE_SHIFT       10       // Shift 10 bits to convert from/to UTF32
#define SURROGATE_HALFBASE    0x10000  // UTF32 -> UTF16 split point for high order
#define SURROGATE_HALFMASK    0x3FF    // UTF32 -> UTF16 mask to get low order
/*-------------------------------------------------------------------------------------------------
 * UTF 32 <-> UTF 16 defines
 *-------------------------------------------------------------------------------------------------*/
#define MAX_UTF16             0xFFFF   // Max Basic Multilingual Plane (BMP) (Highest UTF16 char)
#define MAX_UTF16_32          0x10FFFF // Max 32 bit that can be converted to 16 bit (highest surrogate)



/*-------------------------------------------------------------------------------------------------
 * Convert32to16
 *    Converts UTF32 to UTF16  -- Stolen from ODPPPrvDataTypeS
 *-------------------------------------------------------------------------------------------------*/
int Convert32To16(const unsigned int *Utf32,
                     int Utf32Chars,
                     unsigned short *Utf16,
                     int Utf16Size,
                     int *pOutUtf16Size)
{
   int mErrorCode = RC_SUCCESS;
   unsigned int u32;
   int mInCharsUsed = 0;
   int mOutCharsUsed = 0;

   // Keep getting UTF32 data
   while (Utf32Chars)
   {
      u32 = *Utf32++;
      Utf32Chars--;
      mInCharsUsed++;

      // End of string
      if (!u32)
         break;

      // Will this fit as is?
      if (u32 <= MAX_UTF16)
      {
         // Still, if this is a high or low surrogate it is an error in UTF32
         if (u32 >= SURROGATE_HIGH_START && u32 <= SURROGATE_LOW_END)
         {
            mErrorCode = RC_FATAL; //EINVAL
            break;
         }

         // Ok, fits just find thanks, do it if not length request
         if (Utf16)
         {
            if (Utf16Size < 1)
            {
               mErrorCode = RC_FATAL; //ODPP_DATACNV_ERR_DATA_TRUNCATED;
               break;
            }

            *Utf16++ = (unsigned short)u32;
            Utf16Size--;
         }

         mOutCharsUsed++;
      }

      // Can this be converted to UTF16 with surrogates?
      else if (u32 > MAX_UTF16_32)
      {
         mErrorCode = RC_FATAL; //EINVAL;
         break;
      }

      // Put out surrogate
      else
      {
         // Check it out
         if (Utf16)
         {
            if (Utf16Size < 2)
            {
               mErrorCode = RC_FATAL; //ODPP_DATACNV_ERR_DATA_TRUNCATED;
               break;
            }

            Utf16Size -= 2;

            mOutCharsUsed++;

            // Between 0xFFFF -> 0x10FFF, convert to surrogate UTF16
            u32 -= SURROGATE_HALFBASE;

            // Put out high surrogate
            *Utf16++ = (unsigned short)( (u32 >> SURROGATE_SHIFT) + SURROGATE_HIGH_START );

            // Put out low surrogate
            *Utf16++ = (unsigned short)( (u32 & SURROGATE_HALFMASK) + SURROGATE_LOW_START );
         }

         // Length request
         else
         {
            mOutCharsUsed += 2;
         }
      }
   }

   if (Utf16)
      if (Utf16Size > 0)
         *Utf16 = 0;

   if(NULL != pOutUtf16Size)
      (*pOutUtf16Size) = mOutCharsUsed;

   return mErrorCode == 0 ? RC_SUCCESS : mErrorCode;
}
#endif
#ifndef ENV_VSTG_GET_REGION
#define ENV_VSTG_GET_REGION(pparea, len, pool, option) \
	(*(pparea) = malloc((len)), 0)
#define ENV_VSTG_FREE_REGION(parea, len, pool) ((void)free(parea),0)
#endif

/* This probably belongs in a helper file, rather than in a JNI file */
OptimReportingMessage_t *allocateMessagesArray(int numElements)
{
	OptimReportingMessage_t *allocatedBlock = NULL;
	int rc;
	/* Will return NULL on allocation failure */
	size_t allocSize = numElements * (sizeof(OptimReportingMessage_t) + sizeof(long long)); // Add extra for alignment
	rc = ENV_VSTG_GET_REGION((void **)&allocatedBlock, allocSize, 0, 0);
	if (rc != 0) {
		allocatedBlock = NULL;
	}
	return(allocatedBlock);
}

void freeMessagesArray(OptimReportingMessage_t *allocatedBlock, int numElements)
{
	int rc = 0;
	size_t allocSize = numElements * (sizeof(OptimReportingMessage_t) + sizeof(long long)); // Add extra for alignment
	if (allocatedBlock == NULL) {
		return;
	}
	rc = ENV_VSTG_FREE_REGION(allocatedBlock, allocSize, 0);
	// Not sure what to do if rc != 0.  FAIL
}


void setMessageArrayElement(OptimReportingMessage_t *pBlock, int elementNumber, int id, wchar_t message[ORC_MAX_LEN], int msgType)
{
	if (pBlock == (OptimReportingMessage_t *) NULL) {
		return; //Probably should scream bloody murder here.
	}
	if (message == (wchar_t *) NULL) {
		return; // Again, we should probably complain bitterly
	}
	pBlock[elementNumber].id = id;
	wcscpy(pBlock[elementNumber].message,message);
	pBlock[elementNumber].type = msgType;
	return;
}

/* Now do the same for RequestOptions */
OptimReportingNameValuePair_t *allocateRequestOptionsArray(int numElements)
{
	OptimReportingNameValuePair_t *allocatedBlock = NULL;
	int rc;
	/* Will return NULL on allocation failure */
	size_t allocSize = numElements * (sizeof(OptimReportingNameValuePair_t) + sizeof(long long)); // Add extra for alignment
	rc = ENV_VSTG_GET_REGION((void **)&allocatedBlock, allocSize, 0, 0);
	if (rc != 0) {
		allocatedBlock = NULL;
	}
	return(allocatedBlock);
}

void freeRequestOptionsArray(OptimReportingNameValuePair_t *allocatedBlock, int numElements)
{
	int rc = 0;
	size_t allocSize = numElements * (sizeof(OptimReportingNameValuePair_t) + sizeof(long long)); // Add extra for alignment
	if (allocatedBlock == NULL) {
		return;
	}
	rc = ENV_VSTG_FREE_REGION(allocatedBlock, allocSize, 0);
	// Not sure what to do if rc != 0.  FAIL
}


void setRequestOptionsArrayElement(OptimReportingNameValuePair_t *pBlock, int elementNumber, wchar_t name[ORC_MAX_LEN], wchar_t value[ORC_MAX_LEN])
{
	if (pBlock == (OptimReportingNameValuePair_t *) NULL) {
		return; //Probably should scream bloody murder here.
	}
	if (name == (wchar_t *) NULL) {
		return; // Again, we should probably complain bitterly
	}
	if (value == (wchar_t *) NULL) {
		return; // Again, we should probably complain bitterly
	}
	wcscpy(pBlock[elementNumber].name, name);
	wcscpy(pBlock[elementNumber].value, value);
	return;
}

/* And for ExtendedObjects */
OptimReportingNumberPair_t *allocateExtendedObjectsArray(int numElements)
{
	OptimReportingNumberPair_t *allocatedBlock = NULL;
	int rc;
	/* Will return NULL on allocation failure */
	size_t allocSize = numElements * (sizeof(OptimReportingNumberPair_t) + sizeof(long long)); // Add extra for alignment
	rc = ENV_VSTG_GET_REGION((void **)(&allocatedBlock), allocSize, 0, 0);
	if (rc != 0) {
		allocatedBlock = NULL;
	}
	return(allocatedBlock);
}

void freeExtendedObjectsArray(OptimReportingNumberPair_t *allocatedBlock, int numElements)
{
	int rc = 0;
	size_t allocSize = numElements * (sizeof(OptimReportingNumberPair_t) + sizeof(long long)); // Add extra for alignment
	rc = ENV_VSTG_FREE_REGION(allocatedBlock, allocSize, 0);
	// Not sure what to do if rc != 0.  FAIL
}


void setExtendedObjectsArrayElement(OptimReportingNumberPair_t *pBlock, int elementNumber, int number1, int number2)
{
	if (pBlock == (OptimReportingNumberPair_t *) NULL) {
		return; //Probably should scream bloody murder here.
	}
	pBlock[elementNumber].number1 = number1;
	pBlock[elementNumber].number2 = number2;
	return;
}

/* And for Entities */
OptimReportingEntity_t *allocateEntitiesArray(int numElements)
{
	OptimReportingEntity_t *allocatedBlock = NULL;
	int rc;
	/* Will return NULL on allocation failure */
	size_t allocSize = numElements * (sizeof(OptimReportingEntity_t) + sizeof(long long)); // Add extra for alignment
	rc = ENV_VSTG_GET_REGION((void **)&allocatedBlock, allocSize, 0, 0);
	if (rc != 0) {
		allocatedBlock = NULL;
	}
	return(allocatedBlock);
}

void freeEntitiesArray(OptimReportingEntity_t *allocatedBlock, int numElements)
{
	int rc = 0;
	size_t allocSize = numElements * (sizeof(OptimReportingEntity_t) + sizeof(long long)); // Add extra for alignment
	rc = ENV_VSTG_FREE_REGION(allocatedBlock, allocSize, 0);
	// Not sure what to do if rc != 0.  FAIL
}

/* Now do the same for Impact Analysis */
OptimReportingTableImpAnalysis_t *allocateImpactAnalysisArray(int numElements)
{
	OptimReportingTableImpAnalysis_t *allocatedBlock = NULL;
	int rc;
	/* Will return NULL on allocation failure */
	size_t allocSize = numElements * (sizeof(OptimReportingTableImpAnalysis_t) + sizeof(long long)); // Add extra for alignment
	rc = ENV_VSTG_GET_REGION((void **)&allocatedBlock, allocSize, 0, 0);
	if (rc != 0) {
		allocatedBlock = NULL;
	}
	return(allocatedBlock);
}

void freeImpactAnalysisArray(OptimReportingTableImpAnalysis_t *allocatedBlock, int numElements)
{
	int rc = 0;
	size_t allocSize = numElements * (sizeof(OptimReportingTableImpAnalysis_t) + sizeof(long long)); // Add extra for alignment
	if (allocatedBlock == NULL) {
		return;
	}
	rc = ENV_VSTG_FREE_REGION(allocatedBlock, allocSize, 0);
	// Not sure what to do if rc != 0.  FAIL
}


void setImpactAnalysisArrayElement(OptimReportingTableImpAnalysis_t *pBlock, int elementNumber, wchar_t name[ORC_MAX_LEN], 
                                   wchar_t type[ORC_MAX_LEN],
                                   wchar_t status[ORC_MAX_LEN], long long storageSavings)
{
	if (pBlock == (OptimReportingTableImpAnalysis_t *) NULL) {
		return; //Probably should scream bloody murder here.
	}
	if (name == (wchar_t *) NULL) {
		return; // Again, we should probably complain bitterly
	}
	if (type == (wchar_t *) NULL) {
		return; // Again, we should probably complain bitterly
	}
   if (status == (wchar_t *) NULL) {
		return; // Again, we should probably complain bitterly
	}
	wcscpy(pBlock[elementNumber].name, name);
	wcscpy(pBlock[elementNumber].type, type);
   wcscpy(pBlock[elementNumber].status, status);
   pBlock[elementNumber].storageSavings = storageSavings;
	return;
}

/* Now do the same for Join Control */
OptimReportingJoinControl_t *allocateJoinControlArray(int numElements)
{
	OptimReportingJoinControl_t *allocatedBlock = NULL;
	int rc;
	/* Will return NULL on allocation failure */
	size_t allocSize = numElements * (sizeof(OptimReportingJoinControl_t) + sizeof(long long)); // Add extra for alignment
	rc = ENV_VSTG_GET_REGION((void **)&allocatedBlock, allocSize, 0, 0);
	if (rc != 0) {
		allocatedBlock = NULL;
	}
	return(allocatedBlock);
}

void freeJoinControlArray(OptimReportingJoinControl_t *allocatedBlock, int numElements)
{
	int rc = 0;
	size_t allocSize = numElements * (sizeof(OptimReportingJoinControl_t) + sizeof(long long)); // Add extra for alignment
	if (allocatedBlock == NULL) {
		return;
	}
	rc = ENV_VSTG_FREE_REGION(allocatedBlock, allocSize, 0);
	// Not sure what to do if rc != 0.  FAIL
}


void setJoinControlArrayElement(OptimReportingJoinControl_t *pBlock, int elementNumber, wchar_t tableName[ORC_MAX_LEN], 
                                wchar_t joinedTable[ORC_MAX_LEN])
{
	if (pBlock == (OptimReportingJoinControl_t *) NULL) {
		return; //Probably should scream bloody murder here.
	}
	if (tableName == (wchar_t *) NULL) {
		return; // Again, we should probably complain bitterly
	}
	if (joinedTable == (wchar_t *) NULL) {
		return; // Again, we should probably complain bitterly
	}
	wcscpy(pBlock[elementNumber].tableName, tableName);
	wcscpy(pBlock[elementNumber].joinedTable, joinedTable);
	return;
}


/* Too many elements for a convenience setter on Entity... */
/* And for Relationships */
OptimReportingRelationship_t *allocateRelationshipsArray(int numElements)
{
	OptimReportingRelationship_t *allocatedBlock = NULL;
	int rc;
	/* Will return NULL on allocation failure */
	size_t allocSize = numElements * (sizeof(OptimReportingRelationship_t) + sizeof(long long)); // Add extra for alignment
	rc = ENV_VSTG_GET_REGION((void **)&allocatedBlock, allocSize, 0, 0);
	if (rc != 0) {
		allocatedBlock = NULL;
	}
	return(allocatedBlock);
}

void freeRelationshipsArray(OptimReportingRelationship_t *allocatedBlock, int numElements)
{
	int rc = 0;
	size_t allocSize = numElements * (sizeof(OptimReportingRelationship_t) + sizeof(long long)); // Add extra for alignment
	rc = ENV_VSTG_FREE_REGION(allocatedBlock, allocSize, 0);
	// Not sure what to do if rc != 0.  FAIL
}


void setRelationshipsArrayElement(OptimReportingRelationship_t *pBlock, int elementNumber,
		wchar_t relationshipName[ORC_MAX_LEN], wchar_t parentName[ORC_MAX_LEN], 
      wchar_t childName[ORC_MAX_LEN], int parentAccessType, int parentKeyLimit,
		int childAccessType, int childKeyLimit)
{
	if (pBlock == (OptimReportingRelationship_t *) NULL) {
		return; //Probably should scream bloody murder here.
	}
	if (relationshipName == (wchar_t *) NULL) {
		return; // Again, we should probably complain bitterly
	}
	if (parentName == (wchar_t *) NULL) {
		return; // Again, we should probably complain bitterly
	}
	if (childName == (wchar_t *) NULL) {
		return; // Again, we should probably complain bitterly
	}
	wcscpy(pBlock[elementNumber].relationshipName, relationshipName);
	wcscpy(pBlock[elementNumber].parentName, parentName);
	wcscpy(pBlock[elementNumber].childName, childName);
	pBlock[elementNumber].parentAccessType = parentAccessType;
	pBlock[elementNumber].parentKeyLimit = parentKeyLimit;
	pBlock[elementNumber].childAccessType = childAccessType;
	pBlock[elementNumber].childKeyLimit = childKeyLimit;
	return;
}

/* And for Statistics */
OptimReportingStatistics_t *allocateStatisticsArray(int numElements)
{
	OptimReportingStatistics_t *allocatedBlock = NULL;
	int rc;
	/* Will return NULL on allocation failure */
	size_t allocSize = numElements * (sizeof(OptimReportingStatistics_t) + sizeof(long long)); // Add extra for alignment
	rc = ENV_VSTG_GET_REGION((void **)&allocatedBlock, allocSize, 0, 0);
	if (rc != 0) {
		allocatedBlock = NULL;
	}
	return(allocatedBlock);
}

void freeStatisticsArray(OptimReportingStatistics_t *allocatedBlock, int numElements)
{
	int rc = 0;
	size_t allocSize = numElements * (sizeof(OptimReportingStatistics_t) + sizeof(long long)); // Add extra for alignment
	rc = ENV_VSTG_FREE_REGION(allocatedBlock, allocSize, 0);
	// Not sure what to do if rc != 0.  FAIL
}

void setStatisticsArrayElement(OptimReportingStatistics_t *pBlock, int elementNumber,
		                        long long estimatedRowCount, long long fetchBuffer,
                              long long foreignKeyIndex,  long long foreignKeyNoIndex,
                              long long primaryKeyIndex,  long long primaryKeyNoIndex,
                              long long dbDataPerSecond,  long long dbProcessTime,
                              long long dbRowsFetched, long long dbRowsPerSecond,
                              long long dbRowsWritten, long long dbTimeInAct,
                              long long dbTimeInDB, wchar_t name[ORC_MAX_LEN],
                              wchar_t dbmsVendor[ORC_MAX_LEN], wchar_t dbmsVersion[ORC_MAX_LEN],
                              wchar_t relName[ORC_MAX_LEN], wchar_t relSql[ORC_MAX_LEN],
                              int childStrategy, int columnCount,
                              int databaseConnections, int method,
                              int parentStrategy, int rowLength,
                              int stepNumber, wchar_t connectionMode[ORC_MAX_LEN],
                              wchar_t whereClause[ORC_MAX_LEN], wchar_t* rowid,  
                              int relAccess, int relDb2Cost,
                              int relDb2ScanCost, int relDirection,
                              int relKeyLength, int relKeyPerCur,
                              int relLookupKey, int relType,
                              int dbAccessType,
                              int dbKeysPer, int dbOpenCursor,
                              unsigned char cycle, unsigned char lobs,
                              unsigned char selectUniqueRecord, unsigned char relIndex,
                              wchar_t compTableName[ORC_MAX_LEN], wchar_t compRatio[ORC_MAX_LEN])
{
	if (pBlock == (OptimReportingStatistics_t *) NULL) {
		return; //Probably should scream bloody murder here.
	}
	if (name == (wchar_t *) NULL) {
		return; // Again, we should probably complain bitterly
	}
	if (dbmsVendor == (wchar_t *) NULL) {
		return; // Again, we should probably complain bitterly
	}
	if (dbmsVersion == (wchar_t *) NULL) {
		return; // Again, we should probably complain bitterly
	}
   if (relName == (wchar_t *) NULL) {
		return; // Again, we should probably complain bitterly
	}
   if (relSql == (wchar_t *) NULL) {
		return; // Again, we should probably complain bitterly
	}
   if (connectionMode == (wchar_t *) NULL) {
		return; // Again, we should probably complain bitterly
	}
   if (whereClause == (wchar_t *) NULL) {
		return; // Again, we should probably complain bitterly
	}
   if (rowid == (wchar_t *) NULL) {
		return; // Again, we should probably complain bitterly
	}  
   if (compTableName == (wchar_t *) NULL) {
		return; // Again, we should probably complain bitterly
	}

	pBlock[elementNumber].estimatedRowCount = estimatedRowCount;
   pBlock[elementNumber].fetchBuffer = fetchBuffer;
   pBlock[elementNumber].foreignKeyIndex = foreignKeyIndex;
   pBlock[elementNumber].foreignKeyNoIndex = foreignKeyNoIndex;
   pBlock[elementNumber].primaryKeyIndex = primaryKeyIndex;
   pBlock[elementNumber].primaryKeyNoIndex = primaryKeyNoIndex;
   pBlock[elementNumber].dbDataPerSecond = dbDataPerSecond;
   pBlock[elementNumber].dbProcessTime = dbProcessTime;
   pBlock[elementNumber].dbRowsFetched = dbRowsFetched;
   pBlock[elementNumber].dbRowsPerSecond = dbRowsPerSecond;
   pBlock[elementNumber].dbRowsWritten = dbRowsWritten;
   pBlock[elementNumber].dbTimeInAct = dbTimeInAct;
   pBlock[elementNumber].dbTimeInDB = dbTimeInDB;
   wcscpy(pBlock[elementNumber].name, name);
   wcscpy(pBlock[elementNumber].dbmsVendor, dbmsVendor);
   wcscpy(pBlock[elementNumber].dbmsVersion, dbmsVersion);
   wcscpy(pBlock[elementNumber].relName, relName);
   wcscpy(pBlock[elementNumber].relSql, relSql);
   pBlock[elementNumber].childStrategy = childStrategy;
   pBlock[elementNumber].columnCount = columnCount;
   pBlock[elementNumber].databaseConnections = databaseConnections;
   pBlock[elementNumber].method = method;
   pBlock[elementNumber].parentStrategy = parentStrategy;
   pBlock[elementNumber].rowLength = rowLength;
   pBlock[elementNumber].stepNumber = stepNumber;
   wcscpy(pBlock[elementNumber].connectionMode, connectionMode);
   wcscpy(pBlock[elementNumber].whereClause, whereClause);
   wcscpy(pBlock[elementNumber].rowid, rowid);  
   pBlock[elementNumber].relAccess = relAccess;
   pBlock[elementNumber].relDb2Cost = relDb2Cost;
   pBlock[elementNumber].relDb2ScanCost = relDb2ScanCost;
   pBlock[elementNumber].relDirection = relDirection;
   pBlock[elementNumber].relKeyLength = relKeyLength;
   pBlock[elementNumber].relKeyPerCur = relKeyPerCur;
   pBlock[elementNumber].relLookupKey = relLookupKey;
   pBlock[elementNumber].relType = relType;  
   pBlock[elementNumber].dbAccessType = dbAccessType;
   pBlock[elementNumber].dbKeysPer = dbKeysPer;
   pBlock[elementNumber].dbOpenCursor = dbOpenCursor;
   pBlock[elementNumber].cycle = cycle; 
   pBlock[elementNumber].lobs = lobs;
   pBlock[elementNumber].selectUniqueRecord = selectUniqueRecord;
   pBlock[elementNumber].relIndex = relIndex;
   wcscpy(pBlock[elementNumber].compTableName, compTableName);
   wcscpy(pBlock[elementNumber].compRatio, compRatio);
	return;
}




/* Frees all the internall allocated elements of a context, since they should be allocated with our convenience methods. */
void freeContextInternals(OptimReportingContext_t *context)
{
	freeEntitiesArray(context->pEntities, context->numEntities);
	context->pEntities = (OptimReportingEntity_t *)NULL;
	context->numEntities = 0;

	freeExtendedObjectsArray(context->pExtendedObjects, context->numExtendedObjects);
	context->pExtendedObjects = (OptimReportingNumberPair_t *)NULL;
	context->numExtendedObjects = 0;

	freeMessagesArray(context->pMessages, context->numMessages);
	context->pMessages = (OptimReportingMessage_t *)NULL;
	context->numMessages = 0;

	freeRelationshipsArray(context->pRelationships, context->numRelationships);
	context->pRelationships = (OptimReportingRelationship_t *)NULL;
	context->numRelationships = 0;

	freeRequestOptionsArray(context->pRequestOptions, context->numRequestOptions);
	context->pRequestOptions = (OptimReportingNameValuePair_t *)NULL;
	context->numRequestOptions = 0;

	freeStatisticsArray(context->pStatistics, context->numStatistics);
	context->pStatistics = (OptimReportingStatistics_t *)NULL;
	context->numStatistics = 0;
}
/* Done with Array convenience routines */

#ifdef TEST_MODE

void populateTestContext(OptimReportingContext_t *context)
{
	context->reportFile=L"/tmp/ReportTest.xml";
	context->htmlReportFile=L"/tmp/ReportTest.html";
	context->reportType = ORC_ARCHIVE_REPORT;
	context->optimDirectoryName = L"DIR_ORA112";
	context->optimDirectoryDBMSType = L"Oracle";
	context->optimDirectoryDBMSVersion = L"11.2";
	context->optimDirectoryDBMSVendor = L"Oracle";
	context->optimSolutionName = L"Optim Data Growth";
	context->optimSolutionVersion = L"9.1.0.5";
	context->optimRuntimeName = L"Optim Distributed";
	context->optimRuntimeVersion = L"9.1.0.5";
	context->clientUserId = L"optimuser";
	context->errorCount = 1;
	context->infoCount = 3;
	context->requestName = L"OPTIMUSER_ARCHIVE1";
	context->serverName = L"(local)";
	context->serverUserId = L"optimuser";
	context->timeElapsed = 12345L;
	context->timeStarted = L"06-17-2013 13:24:00";
	context->timeFinished = L"06-17-2013 13:27:00";
	context->warningCount = 0;
	/* First instance of an array type of thing... */
	context->numMessages = 2;
	/* Not checking for NULL from allocation... this is test/demo code, not production */
	context->messages = allocateMessagesArray(context->numMessages);
	setMessageArrayElement(context->messages, 0, 0, L"For those about to rock", ORM_MESSAGE_TYPE_ERROR);
	setMessageArrayElement(context->messages, 1, 1, L"We salute you", ORM_MESSAGE_TYPE_WARNING);
	context->totalRows = 90;
	context->totalRowsFirstPass = 30;
	context->totalRowsInError = 1;
	context->totalTables = 2;
	context->numRequestOptions = 1;
	context->requestOptions = allocateRequestOptionsArray(context->numRequestOptions);
	setRequestOptionsArrayElement(context->requestOptions, 0, L"RequestOption 1", L"Splunge");
	context->extractFile = L"/opt/IBM/InfoSphere/Optim/ArchiveFile.af";
	context->accessDefinitionName = L"OPTIMUSER_AD";
	context->accessDefinitionType = ORC_ACCESS_DEFINITION_TYPE;
	context->accessDefinitionModificationDate = L"2013-03-24";
	context->numExtendedObjects = 1;
	context->extendedObjects = allocateExtendedObjectsArray(context->numExtendedObjects);
	setExtendedObjectsArrayElement(context->extendedObjects, 0, 0, ORC_ALIAS_TYPE);
	context->numEntities = 2;
	context->pEntities = allocateEntitiesArray(context->numEntities);
	/* Set everything for 1st entity */
	context->pEntities[0].failed = 0;
	context->pEntities[0].name = L"OPTIM.OPTIM_PARENT";
	context->pEntities[0].recordCount = 30L;
	context->pEntities[0].skipped = 0;
	context->pEntities[0].status = L" ";
	context->pEntities[0].type = ORE_TABLE_TYPE;
	context->pEntities[0].extractAverageRowLength = 255.4;
	context->pEntities[0].extractLob = TRUE;
	context->pEntities[0].extractMaxLobLength = 123456789.0;
	context->pEntities[0].extractCompressionEnabled = FALSE;
	context->pEntities[0].extractCompressionValue = 0;
			/* Next Entity */
	context->pEntities[1].failed = 3;
	context->pEntities[1].name = L"OPTIM.OPTIM_CHILD";
	context->pEntities[1].recordCount = 40L;
	context->pEntities[1].skipped = 0;
	context->pEntities[1].status = L"A-OK";
	context->pEntities[1].type = ORE_TABLE_TYPE;
	context->pEntities[1].extractAverageRowLength = 255.4;
	context->pEntities[1].extractLob = TRUE;
	context->pEntities[1].extractMaxLobLength = 123456789.0;
	context->pEntities[1].extractCompressionEnabled = TRUE;
	context->pEntities[1].extractCompressionValue = 3;
	/* Relationships */
	context->numRelationships = 1;
	context->pRelationships = allocateRelationshipsArray(context->numRelationships);
	setRelationshipsArrayElement(context->pRelationships, 0, L"OPTIM.MOTHER_AND_CHILD_REUNION",
			L"OPTIM.OPTIM_PARENT", L"OPTIM.OPTIM_CHILD", ORR_SCAN_STRATEGY, 0, ORR_SCAN_STRATEGY, 0);
	/* Statistics */
	context->numStatistics = 1;
	context->pStatistics = allocateStatisticsArray(context->numStatistics);
	context->pStatistics[0].childStrategy = ORS_NO_KEYS_STRATEGY;
	context->pStatistics[0].columnCount = 0;
	context->pStatistics[0].cycle = TRUE;
	context->pStatistics[0].dbAccessType = ORS_SCAN;
	context->pStatistics[0].dbDataPerSecond = 0;
	context->pStatistics[0].dbKeysPer = 0;
	context->pStatistics[0].dbOpenCursor = 0;
	context->pStatistics[0].dbProcessTime = 12345L;
	context->pStatistics[0].dbRowsFetched = 1457;
	context->pStatistics[0].dbRowsPerSecond = 1450;
	context->pStatistics[0].dbRowsWritten = 1457;
	context->pStatistics[0].dbTimeInAct = 12300L;
	context->pStatistics[0].dbTimeInDB = 45L;
	context->pStatistics[0].estimatedRowCount = 0;
	context->pStatistics[0].fetchBuffer = 0;
	context->pStatistics[0].foreignKeyIndex = 0;
	context->pStatistics[0].foreignKeyNoIndex = 0;
	context->pStatistics[0].lobs = TRUE;
	context->pStatistics[0].method = ORS_PRIMARY_KEY_METHOD;
	context->pStatistics[0].name = L"Statistic Name";
	context->pStatistics[0].parentStrategy = ORS_NO_KEYS_STRATEGY;
	context->pStatistics[0].primaryKeyIndex = 0;
	context->pStatistics[0].primaryKeyNoIndex = 0;
	context->pStatistics[0].relAccess = ORS_FORCING_ACCESS;
	context->pStatistics[0].relDb2Cost = 0;
	context->pStatistics[0].relDb2ScanCost = 0;
	context->pStatistics[0].relDirection = ORS_PARENT_DIRECTION;
	context->pStatistics[0].relIndex = TRUE;
	context->pStatistics[0].relKeyLength = 0;
	context->pStatistics[0].relKeyPerCur = 0;
	context->pStatistics[0].relLookupKey = 0;
	context->pStatistics[0].relName = L"OPTIM.MOTHER_AND_CHILD_REUNION";
	context->pStatistics[0].relSql = L"SELECT count(FOO) FROM OPTIM.OPTIM_CUSTOMERS";
	context->pStatistics[0].relType = ORS_REL_TYPE_DATABASE;
	context->pStatistics[0].rowLength = 0;
	context->pStatistics[0].selectUniqueRecord = TRUE;
	context->pStatistics[0].stepNumber = 0;
	context->pStatistics[0].dbmsVendor = L"Oracle";
	context->pStatistics[0].dbmsVersion = L"11.2";
}

int
main(int argc, char **argv)
{
	OptimReportingContext_t context;
	int rc = RC_SUCCESS;

	(void)printf("Starting report test\n");
	/* Populate C Context */
	/* Use /tmp/ReportTest.xml and /tmp/ReportTest.html */
	populateTestContext(&context);
	(void)printf("Populated Test context\n");

	/* Call createReport */
	(void)printf("Outputting Report\n");
   
	rc = outputReport(&context);
	(void)printf("Return Code %d, now freeing internals\n", rc);
	/* Deallocate all */
	freeContextInternals(&context);
	(void)printf("Terminating...\n");
	/* End */
	return rc;
}

#endif /* TEST_MODE */
